from datetime import datetime, timedelta
from unittest.mock import MagicMock, patch
import pytest

from server.pdu import PDUState, PDUPort, PDU


def test_PDUState_UNKNOW_is_invalid_action():
    assert PDUState.UNKNOWN not in PDUState.valid_actions()
    assert PDUState.UNKNOWN.is_valid_action is False


def test_PDUState_valid_actions_contain_basics():
    for action in ["ON", "OFF", "REBOOT"]:
        assert action in [s.name for s in PDUState.valid_actions()]
        assert getattr(PDUState, action).is_valid_action is True


def __pdu_port_mock():
    pdu = MagicMock(reserved_port_ids=[])
    pdu.name = "MyPDU"
    port = PDUPort(pdu, 42, label="My Port")

    return pdu, port


def test_PDUPort__defaults():
    pdu, port = __pdu_port_mock()

    assert port.label == "My Port"
    assert repr(port) == "<PDU Port: PDU=MyPDU, ID=42, label=My Port, min_off_time=5>"

    assert port.last_shutdown is None
    assert port.last_known_state == PDUState.UNKNOWN

    pdu.set_port_state.assert_not_called()
    pdu.get_port_state.assert_not_called()


def test_PDUPort__get_state_UNKNOWN_to_OFF():
    pdu, port = __pdu_port_mock()

    port.last_shutdown = None
    port.last_known_state = PDUState.UNKNOWN

    pdu.get_port_state.return_value = PDUState.OFF
    assert port.state == PDUState.OFF

    pdu.get_port_state.assert_called_with(42)
    assert datetime.now() - port.last_shutdown < timedelta(seconds=1)
    assert port.last_known_state == PDUState.OFF


def test_PDUPort__get_state_UNKNOWN_to_ON():
    pdu, port = __pdu_port_mock()

    port.last_shutdown = None
    port.last_known_state = PDUState.UNKNOWN

    pdu.get_port_state.return_value = PDUState.ON
    assert port.state == PDUState.ON

    assert port.last_shutdown is None
    assert port.last_known_state == PDUState.ON


def test_PDUPort__get_state_ON_to_OFF():
    pdu, port = __pdu_port_mock()

    port.last_shutdown = None
    port.last_known_state = PDUState.ON

    pdu.get_port_state.return_value = PDUState.OFF
    assert port.state == PDUState.OFF

    assert port.last_shutdown is not None
    assert port.last_known_state == PDUState.OFF


def test_PDUPort__get_state_OFF_to_OFF():
    pdu, port = __pdu_port_mock()

    port.last_shutdown = wanted_last_shutdown = datetime.now()
    port.last_known_state = PDUState.OFF

    pdu.get_port_state.return_value = PDUState.OFF
    assert port.state == PDUState.OFF

    pdu.get_port_state.assert_called_with(42)
    assert port.last_shutdown == wanted_last_shutdown
    assert port.last_known_state == PDUState.OFF


@patch("time.sleep")
def test_PDUPort__setting_OFF_to_ON(sleep_mock):
    pdu, port = __pdu_port_mock()

    port.last_shutdown = wanted_last_shutdown = datetime.now()
    port.last_known_state = PDUState.OFF

    pdu.get_port_state.return_value = PDUState.OFF
    sleep_mock.assert_not_called()
    port.set(PDUState.ON)

    # Make sure the minimum off time was respected
    sleep_mock.assert_called_once()
    sleep_time = sleep_mock.call_args[0][0]
    assert (port.min_off_time - sleep_time) < 1.0

    pdu.set_port_state.assert_called_with(port.port_id, PDUState.ON)
    assert pdu.set_port_state.call_count == 1
    assert port.last_shutdown == wanted_last_shutdown
    assert port.last_known_state == PDUState.ON


@patch("time.sleep")
def test_PDUPort__setting_ON_to_OFF(sleep_mock):
    pdu, port = __pdu_port_mock()

    port.last_shutdown = None
    port.last_known_state = PDUState.ON

    pdu.get_port_state.return_value = PDUState.ON
    port.set(PDUState.OFF)

    sleep_mock.assert_not_called()
    pdu.set_port_state.assert_called_with(port.port_id, PDUState.OFF)
    assert pdu.set_port_state.call_count == 1
    assert port.last_shutdown is not None
    assert port.last_known_state == PDUState.OFF


@patch("time.sleep")
def test_PDUPort__setting_OFF_to_OFF(sleep_mock):
    pdu, port = __pdu_port_mock()

    port.last_shutdown = wanted_last_shutdown = datetime.now()
    port.last_known_state = PDUState.OFF

    pdu.get_port_state.return_value = PDUState.OFF
    port.set(PDUState.OFF)

    sleep_mock.assert_not_called()
    assert port.last_shutdown == wanted_last_shutdown
    assert port.last_known_state == PDUState.OFF


def test_PDUPort_eq():
    params = {
        "pdu": "pdu",
        "port_id": 42,
        "label": "label",
        "min_off_time": "min_off_time"
    }

    assert PDUPort(**params) == PDUPort(**params)
    for param in params:
        n_params = dict(params)
        n_params[param] = "modified"
        assert PDUPort(**params) != PDUPort(**n_params)


def test_PDU_defaults():
    config = {"key": "val"}
    pdu = PDU("MyPDU", config)

    assert pdu.name == "MyPDU"
    assert pdu.config == config
    assert pdu.ports == []
    assert pdu.set_port_state(42, PDUState.ON) is False
    assert pdu.get_port_state(42) == PDUState.UNKNOWN
    assert pdu.default_min_off_time == 30


def test_PDU_supported_pdus():
    pdus = PDU.supported_pdus()
    assert "dummy" in pdus


def test_PDU_create():
    pdu = PDU.create("dummy", "name", {})
    assert pdu.name == "name"

    with pytest.raises(ValueError):
        PDU.create("invalid", "name", {})


def test_PDU_reserved_port0():
    pdu = PDU.create("dummy", "name", {"ports": ['0', '1', '2', '3']}, reserved_port_ids=[1])
    assert pdu.ports[1].reserved is True
    assert pdu.get_port_state(1) == PDUState.ON
    pdu.ports[1].set(PDUState.OFF)
    assert pdu.get_port_state(1) == PDUState.ON


def test_PDU_reserved_unreserved_port():
    pdu = PDU.create("dummy", "name", {"ports": ['0', '1', '2', '3', '4', '5']}, ['3', '4'])
    assert pdu.ports[3].reserved is True
    assert pdu.ports[4].reserved is True
    assert pdu.reserved_port_ids == {'3', '4'}

    pdu.unreserve_port('3')
    assert pdu.ports[3].reserved is False
    assert pdu.ports[4].reserved is True
    assert pdu.reserved_port_ids == {'4'}

    # Make sure nothing goes wrong when removing an already-removed port
    pdu.unreserve_port('3')

    pdu.reserve_port('5')
    assert pdu.ports[4].reserved is True
    assert pdu.ports[5].reserved is True
    assert pdu.reserved_port_ids == {'4', '5'}



from dataclasses import dataclass
from jinja2 import Template
from typing import List
import requests
import math
import platform
import psutil

from .logger import logger
from . import config


class SanitizedFieldsMixin:
    @classmethod
    def from_api(cls, fields, **kwargs):
        valid_fields = {f.name for f in cls.__dataclass_fields__.values()}

        sanitized_kwargs = dict(fields)
        for arg in fields:
            if arg not in valid_fields:
                sanitized_kwargs.pop(arg)

        return cls(**sanitized_kwargs, **kwargs)


@dataclass
class GitlabRunnerRegistration(SanitizedFieldsMixin):
    id: int
    token: str


@dataclass
class Runner(SanitizedFieldsMixin):
    id: int
    description: str
    active: bool
    paused: bool
    online: bool
    tag_list: List[str]
    maximum_timeout: int


def register_runner(gitlab_url: str, registration_token: str,
                    description: str, tag_list: list[str],
                    run_untagged: bool = False, maximum_timeout: int = 3600,
                    runner_type: str = "instance_type",
                    group_id: int = None, project_id: int = None):
    # Make sure the group or project id is set when needed by the runner type
    assert runner_type in {"instance_type", "group_type", "project_type"}
    assert runner_type != "group_type" or group_id is not None
    assert runner_type != "project_type" or project_id is not None

    params = {
        "description": description,
        "tag_list": ",".join(tag_list),
        "run_untagged": run_untagged,
        "maximum_timeout": maximum_timeout,
        "runner_type": runner_type,
    }

    if group_id:
        params["group_id"] = group_id
    if project_id:
        params["project_id"] = project_id

    r = requests.post(f"{gitlab_url}/api/v4/user/runners",
                      headers={"PRIVATE-TOKEN": registration_token}, params=params)
    if r.status_code == 201:
        return GitlabRunnerRegistration.from_api(r.json())
    else:
        return None


def unregister_runner(gitlab_url: str, token: str):
    r = requests.delete(f"{gitlab_url}/api/v4/runners", params={"token": token})
    return r.status_code == 204


def verify_runner_token(gitlab_url: str, token: str):
    # WARNING: The interface for this function is so that we will return
    # False *ONLY* when Gitlab tells us the token is invalid.
    # If Gitlab is unreachable, we will return True.
    #
    # This is a conscious decision, as we never want to throw away a perfectly-good
    # token, just because of a network outtage.

    r = requests.post(f"{gitlab_url}/api/v4/runners/verify", params={"token": token})
    return not r.status_code == 403


def runner_details(gitlab_url: str, private_token: str, runner_id: int):
    r = requests.get(f"{gitlab_url}/api/v4/runners/{runner_id}",
                     headers={"PRIVATE-TOKEN": private_token})
    if r.status_code == 200:
        return Runner.from_api(r.json())
    else:
        return None


def generate_gateway_runner_tags():
    def next_power_of_2(x):
        return 1 if x == 0 else 2**math.ceil(math.log2(x))

    tags = {f"{config.FARM_NAME}-gateway", 'CI-gateway', f"cpu:arch:{platform.machine()}"}

    cpu_count = psutil.cpu_count(logical=False)
    tags.add(f"cpu:cores:{cpu_count}")
    if cpu_count >= 8:
        tags.add("cpu:cores:8+")
    if cpu_count >= 16:
        tags.add("cpu:cores:16+")
    if cpu_count >= 32:
        tags.add("cpu:cores:32+")

    mem_gib = next_power_of_2(psutil.virtual_memory().total / 1024 / 1024 / 1024)
    tags.add(f"mem:size:{mem_gib}GiB")
    if mem_gib >= 8:
        tags.add("mem:size:8+GiB")
    if mem_gib >= 32:
        tags.add("mem:size:32+GiB")

    return tags


def generate_runner_config(mars_db):
    logger.info("Generate the GitLab runner configuration")
    with open(config.GITLAB_CONF_TEMPLATE_FILE) as f:
        params = {
            "config": config,
            "mars_db": mars_db,
            "cpu_count": psutil.cpu_count(),
            "ram_total_MB": psutil.virtual_memory().total / 1e6
        }
        config_toml = Template(f.read()).render(**params)

    with open(config.GITLAB_CONF_FILE, 'w') as f:
        f.write(config_toml)



import os
import socket
from typing import Dict

BASE_DIR = os.path.dirname(__file__)


def template(filename):
    return os.path.join(os.path.join(BASE_DIR, 'templates'), filename)


def job_template(filename):
    return os.path.join(os.path.join(BASE_DIR, 'job_templates'), filename)


def get_farm_name_from_hostname():
    hostname = socket.gethostname()

    # Recognize the format: $FARM_NAME-gateway
    farm_name = hostname.removesuffix('-gateway')
    if hostname != farm_name:
        return farm_name

    return None


def as_boolean(key):
    value = str(globals().get(key, "")).strip().lower()

    if value in ["true", "enabled", "1"]:
        return True
    elif value in ["false", "disabled", "0"]:
        return False
    else:
        raise ValueError((f"The value '{value}' is not an accepted boolean. "
                          "Accepted values: true/false, enabled/disabled, 1/0"))


# Note: Don't forget to update /documentation/docs/executor.rst
configurables = {
    'CONSOLE_PATTERN_DEFAULT_MACHINE_UNFIT_FOR_SERVICE_REGEX': None,
    'EXECUTOR_URL': 'http://ci-gateway',
    'EXECUTOR_HOST': '0.0.0.0',
    'EXECUTOR_PORT': 80,
    'EXECUTOR_HTTP_IPv4_SOCKET_NAME': 'http_ipv4',
    'EXECUTOR_REGISTRATION_JOB': job_template('register.yml.j2'),
    'EXECUTOR_BOOTLOOP_JOB': job_template('bootloop.yml.j2'),
    'EXECUTOR_VPDU_ENDPOINT': None,
    'EXECUTOR_ARTIFACT_CACHE_ROOT': None,
    'SERGENT_HARTMAN_BOOT_COUNT': '100',
    'SERGENT_HARTMAN_QUALIFYING_BOOT_COUNT': '100',
    'SERGENT_HARTMAN_REGISTRATION_RETRIAL_DELAY': '120',
    'SERGENT_HARTMAN_QUICK_CHECK': 'enabled',
    'GITLAB_CONF_FILE': '/etc/gitlab-runner/config.toml',
    'GITLAB_CONF_TEMPLATE_FILE': template('gitlab_runner_config.toml.j2'),
    'FARM_NAME': get_farm_name_from_hostname(),
    'MARS_DB_FILE': '/config/mars_db.yaml',
    'BOOTS_DB_USER_FILE': '/config/boots_db.yml.j2',
    'BOOTS_DB_FILE': template('boots_db.yml.j2'),
    'SALAD_URL': 'http://ci-gateway:8005',
    'BOOTS_TFTP_ROOT': '/cache/boots/tftp',
    'BOOTS_DISABLE_SERVERS': None,
    'MINIO_URL': 'http://ci-gateway:9000',
    'MINIO_ROOT_USER': 'minioadmin',
    'MINIO_ROOT_PASSWORD': 'minio-root-password',
    'MINIO_ADMIN_ALIAS': 'local',
    'PRIVATE_INTERFACE': 'private',
    'BOOTS_DEFAULT_KERNEL': 'http://ci-gateway:9000/boot/default_kernel',
    'BOOTS_DEFAULT_INITRD': 'http://ci-gateway:9000/boot/default_boot2container.cpio.xz',
    'BOOTS_DEFAULT_CMDLINE': 'b2c.container="-ti --tls-verify=false docker://ci-gateway:8002/gfx-ci/ci-tron/machine-registration:latest register" b2c.ntp_peer="ci-gateway" b2c.cache_device=none loglevel=6',  # noqa
    'BOOTS_DHCP_IPv4_SOCKET_NAME': 'dhcp_ipv4',
    'BOOTS_TFTP_IPv4_SOCKET_NAME': 'tftp_ipv4',
}
# Note: Don't forget to update /documentation/docs/executor.rst

__all__ = []


for config_option, default in configurables.items():
    globals()[config_option] = os.environ.get(config_option,
                                              default)
    __all__.append(config_option)


def job_environment_vars() -> Dict[str, str]:  # pragma: nocover
    """Return environment variables useful for job submission as a
    dictionary."""

    ret = {
        'MINIO_URL': globals()['MINIO_URL'],
    }

    for var, val in os.environ.items():
        if var.startswith('EXECUTOR_JOB__'):
            ret[var.lstrip('EXECUTOR_JOB__')] = val

    return ret



from unittest.mock import MagicMock
import pytest
import time

from server.pdu.drivers.apc import ApcMasterswitchPDU


@pytest.fixture(autouse=True)
def reset_easysnmp_mock(monkeypatch):
    import server.pdu.drivers.snmp as snmp

    global Session, time_sleep
    m1, m2 = MagicMock(), MagicMock()
    # REVIEW: I wonder if there's a clever way of covering the
    # difference in import locations between here and snmp.py
    monkeypatch.setattr(snmp, "Session", m1)
    monkeypatch.setattr(time, "sleep", m2)
    Session = m1
    time_sleep = m2


def test_driver_ApcMasterswitchPDU_check_OIDs():
    pdu = ApcMasterswitchPDU("MyPDU", {"hostname": "127.0.0.1"})

    assert pdu.outlet_labels_oid == f"{pdu.oid_enterprise}.318.1.1.4.4.2.1.4"
    assert pdu.outlet_status_oid(10) == f"{pdu.oid_enterprise}.318.1.1.4.4.2.1.3.10"
    assert pdu.outlet_ctrl_oid(10) == f"{pdu.oid_enterprise}.318.1.1.4.4.2.1.3.10"



from unittest.mock import patch
import pytest

import server.config as config


@patch("server.config.socket.gethostname")
def test_get_farm_name_from_hostname(gethostname_mock):
    gethostname_mock.return_value = "unformated_name"
    assert config.get_farm_name_from_hostname() is None

    gethostname_mock.return_value = "testfarm-gateway"
    assert config.get_farm_name_from_hostname() == "testfarm"


def test_as_boolean():
    try:
        # Check for the conditions to be true
        for value in [' eNaBlEd  ', ' tRuE ', "1 ", 1, True]:
            config.TEST_AS_BOOLEAN = value
            assert config.as_boolean("TEST_AS_BOOLEAN")

        # Check for the conditions to be true
        for value in [' dIsAbLeD  ', ' FaLsE ', "  0 ", 0, False]:
            config.TEST_AS_BOOLEAN = value
            assert not config.as_boolean("TEST_AS_BOOLEAN")

        # Check that any other value is considered invalid
        for value in [None, "data", 42, ""]:
            config.TEST_AS_BOOLEAN = value

            with pytest.raises(ValueError) as exc:
                config.as_boolean("TEST_AS_BOOLEAN")

            assert "is not an accepted boolean" in str(exc)
    finally:
        del config.TEST_AS_BOOLEAN



from dataclasses import asdict
from functools import cached_property
from pathlib import Path
from typing import List, Optional
import re

from pydantic import NonNegativeInt, DirectoryPath, field_validator, model_validator
from pydantic.dataclasses import dataclass

from .. import PDU, PDUPort, PDUState


class USBHubDevice:
    def __init__(self, base_path):
        self.base_path = Path(base_path).resolve()

        if self.bDeviceClass != 0x09:
            raise ValueError(f"The USB device at location '{self.base_path}' is not a USB hub")

        self.ports = [
            str(oid) for oid in [i + 1 for i in range(self.maxchild)] if self.port_path(oid).exists()
        ]

    def __read(self, filename):
        return (self.base_path / filename).read_text().strip()

    def __read_int(self, filename, base=10):
        if v := self.__read(filename):
            return int(v, base)

    @cached_property
    def devpath(self):
        return self.__read("devpath")

    @cached_property
    def serial(self):
        return self.__read("serial")

    @cached_property
    def maxchild(self):
        return self.__read_int("maxchild")

    @cached_property
    def idVendor(self):
        return self.__read_int("idVendor", 16)

    @cached_property
    def idProduct(self):
        return self.__read_int("idProduct", 16)

    @cached_property
    def bDeviceClass(self):
        return self.__read_int("bDeviceClass", 16)

    @cached_property
    def speed(self):
        return self.__read_int("speed")

    @cached_property
    def controller_path(self):
        return self.base_path.parent.parent

    def port_path(self, port_id):
        paths = list(self.base_path.glob(f"*:1.0/*-port{port_id}/disable"))

        if len(paths) > 1:
            msg = f"Found more than one candidate port for port_id {port_id} in the hub at location '{self.base_path}'"
            raise ValueError(msg)
        elif len(paths) == 1:
            return paths[0]
        else:
            return Path()

    def set_port_state(self, port_id, state):
        if state not in [PDUState.OFF, PDUState.ON]:
            raise ValueError(f"Unsupported state {state.name}")

        self.port_path(port_id).write_text("0\n" if state == PDUState.ON else "1\n")

    def get_port_state(self, port_id):
        state = self.port_path(port_id).read_text().strip()
        return PDUState.ON if state == "0" else PDUState.OFF


@dataclass
class USBHubDeviceMatchConfig:
    devpath: str
    idVendor: NonNegativeInt
    idProduct: NonNegativeInt

    maxchild: Optional[NonNegativeInt] = None
    speed: Optional[NonNegativeInt] = None

    @field_validator("devpath")
    @classmethod
    def devpath_has_expected_format(cls, v):
        if not re.match(r"^(\d+\.?)*\d+$", v):
            raise ValueError("The devpath is not a dot-separated list of integers")
        return v

    def matches(self, hub):
        for field_name in asdict(self):
            if value := getattr(self, field_name):
                if getattr(hub, field_name) != value:
                    return False
        return True


@dataclass
class USBHubPDUConfig:
    controller: Optional[DirectoryPath] = None
    devices: Optional[List[USBHubDeviceMatchConfig]] = None

    serial: Optional[str] = None

    # Default paths
    USB_DEVICES_PATH = Path("/sys/bus/usb/devices")

    @field_validator("serial")
    @classmethod
    def serial_is_stripped(cls, v):
        return v.strip()

    @model_validator(mode="after")
    def ensure_we_have_the_right_parameters(self) -> "USBHubPDUConfig":
        if self.serial is not None:
            return self
        elif self.controller is not None and self.devices is not None:
            if len(self.devices) == 0:
                raise ValueError("At least one device should be set")
            return self
        else:
            raise ValueError("Neither a `serial` nor the tuple (`controller`, `devices`) was found in the config")

    @property
    def hubs(self):
        hubs = []
        if self.serial:
            for serial_path in self.USB_DEVICES_PATH.glob("*/serial"):
                if serial_path.read_text().strip() == self.serial:
                    try:
                        hubs.append(USBHubDevice(serial_path.parent))
                    except Exception:  # pragma: nocover
                        continue

            if len(hubs) == 0:
                raise ValueError(f"No USB Hubs with serial '{self.serial}' found")
        elif self.controller and self.devices:
            for dev_cfg in self.devices:
                dev_cfg_hubs = []
                for dev in self.controller.glob(f"usb*/*-{dev_cfg.devpath}"):
                    try:
                        hub = USBHubDevice(dev)
                    except Exception:
                        continue

                    if dev_cfg.matches(hub):
                        dev_cfg_hubs.append(hub)
                    else:
                        print(f"{dev_cfg}.matches({hub}) = False")
                if len(dev_cfg_hubs) == 0:
                    raise ValueError(f"Could not find a USB device matching {dev_cfg}")
                elif len(dev_cfg_hubs) == 1:
                    hubs.extend(dev_cfg_hubs)
                else:
                    raise ValueError(f"Found more than one USB device match {dev_cfg}")

        return hubs


class USBHubPDU(PDU):
    @cached_property
    def associated_hubs(self):
        hubs = USBHubPDUConfig(**self.config).hubs
        if len(hubs) == 0:
            return []

        # Use the first hub as a reference
        ref = hubs[0]

        # Ensure that all the hubs share the same controller (needed for serial-based devices)
        if not all([h.controller_path == ref.controller_path for h in hubs]):
            raise ValueError("Not all hubs are connected to the same USB controller")

        # Ensure that all the hubs have the same ports
        if not all([h.ports == ref.ports for h in hubs]):
            raise ValueError("Not all hubs agree on the list of ports")

        # Ensure that all the hubs have different speeds like they should if a controller had
        # support for multiple USB versions, but all connected to the same USB lines
        speeds = {h.speed for h in hubs}
        if len(speeds) != len(hubs):
            raise ValueError("Some hubs unexpectedly share the same speed")

        return hubs

    def __init__(self, name, config, reserved_port_ids=[]):
        super().__init__(name, config, reserved_port_ids)

        if len(list(self.associated_hubs)) == 0:
            raise ValueError("No associated hub devices found")

        self._ports = [
            PDUPort(self, str(oid))
            for oid in self.associated_hubs[0].ports
        ]

    @property
    def ports(self):
        return self._ports

    def set_port_state(self, port_id, state):
        for h in self.associated_hubs:
            h.set_port_state(port_id, state)

    def get_port_state(self, port_id):
        states = {h.get_port_state(port_id) for h in self.associated_hubs}

        if len(states) == 1:
            return states.pop()
        else:
            return PDUState.UNKNOWN



from functools import cached_property

import logging
import traceback
import os

import usb.core
import usb.util

"""
Interact with fastboot-enabled devices

See https://android.googlesource.com/platform/system/core/+/master/fastboot/README.md
for an overview of the protocol.
"""


class FastbootDevice:
    # See https://android.googlesource.com/platform/system/core/+/eclair-release/fastboot/fastboot.c#156
    FASTBOOT_INTERFACE_CLASS = 0xff
    FASTBOOT_INTERFACE_SUBCLASS = 0x42
    FASTBOOT_INTERFACE_PROTOCOL = 0x03

    def __init__(self, device: usb.core.Device, write_chunk=1024*1024):
        self.device = device
        self.write_chunk = write_chunk

        self.ep_read = self.ep_write = None
        for ep in self.intf:
            if usb.util.endpoint_type(ep.bmAttributes) == usb.util.ENDPOINT_TYPE_BULK:
                if usb.util.endpoint_direction(ep.bEndpointAddress) == usb.util.ENDPOINT_IN:
                    self.ep_read = ep
                else:
                    self.ep_write = ep

        assert self.ep_read
        assert self.ep_write

    def release(self):
        self.device._ctx.release_all_interfaces(self.device)

    @cached_property
    def intf(self):
        # Go through all the configurations to find a fastboot interface
        for cfg in self.device:
            intf = usb.util.find_descriptor(cfg,
                                            bInterfaceClass=self.FASTBOOT_INTERFACE_CLASS,
                                            bInterfaceSubClass=self.FASTBOOT_INTERFACE_SUBCLASS,
                                            bInterfaceProtocol=self.FASTBOOT_INTERFACE_PROTOCOL)
            if intf is not None:
                return intf
            else:  # pragma: nocover
                raise ValueError("The device is not a fastboot-compatible")

    def read(self, length, timeout=None):
        return self.ep_read.read(length, timeout=timeout)

    def write(self, data, timeout=None):
        for start in range(0, len(data), self.write_chunk):
            self.ep_write.write(data[start:start+self.write_chunk], timeout=timeout)

    def read_response(self, timeout=None):
        line = self.read(64, timeout=timeout)
        header = bytes(line[:4])
        data = bytes(line[4:])
        return header, data

    def run_cmd(self, cmd, arg=None, timeout=None, wanted_status=b"OKAY"):
        if not isinstance(cmd, bytes):
            cmd = cmd.encode()

        if arg and not isinstance(arg, bytes):
            arg = arg.encode()

        full_cmd = cmd + b":" + arg if arg else cmd
        self.write(full_cmd)

        lines = []
        while True:
            header, data = self.read_response(timeout=timeout)

            lines.append((header, data))

            # Detect the end of the line
            if header in {b'OKAY', b'DATA'}:
                break
            elif header not in {b'INFO'}:  # On failure
                raise ValueError(f"The command '{full_cmd}' failed ({header}) with error '{data}'")

        if wanted_status and header != wanted_status:
            msg = f"The command '{full_cmd}' returned the status '{header}' instead of the expected '{wanted_status}'"
            raise ValueError(msg)

        return lines

    def getvar(self, name):
        if not isinstance(name, bytes):
            name = name.encode()

        # Get all the listed variables, ignoring the last line which simply
        # contains the execution summary
        variables = dict()
        for _, data in self.run_cmd(b'getvar', name)[0:-1]:
            # Look for the last ':'
            if idx := bytes.rfind(data, b':'):
                name = data[0:idx].decode()
                value = data[idx+1:].decode()
                variables[name] = value

        if len(variables) > 1:
            return variables
        elif len(variables) == 1:
            return list(variables.values())[0]
        else:
            return None

    @cached_property
    def variables(self):
        return self.getvar(b"all")

    def upload(self, fp):
        # Check the size
        pos = fp.tell()
        fp.seek(0, os.SEEK_END)
        size = fp.tell()
        fp.seek(pos, os.SEEK_SET)

        # Setup the transfer
        resp = self.run_cmd(b'download:' + f"{size:08x}".encode(), wanted_status=b"DATA")
        assert int(resp[-1][1].decode().rstrip('\x00'), 16) == size

        # Send the file
        while True:
            chunk = fp.read(self.write_chunk)
            if len(chunk) == 0:
                break

            self.ep_write.write(chunk)

        # Wait for the OK
        header, data = self.read_response()
        assert header == b"OKAY"

    def boot(self):
        self.run_cmd(b'boot', timeout=15000)

    @classmethod
    def from_serial(cls, serialno, write_chunk=1024*1024):
        for device in usb.core.find(find_all=True):
            try:
                if device.serial_number != serialno:
                    continue
            except Exception:
                # Ignore errors related to reading the serial number as they
                # are likely related to permission issues and thus not a device
                # we should be caring about
                continue

            return cls(device, write_chunk=write_chunk)

        raise ValueError(f"Couldn't find a device with serialno '{serialno}'")


class Fastbootd:
    def __init__(self, logger=None):
        self.logger = logger

        # setup logger
        if self.logger is None:
            self.logger = logging.getLogger('FASTBOOT')
            self.logger.propagate = False
            handler = logging.StreamHandler()
            formatter = logging.Formatter('[%(levelname)s] %(message)s')
            handler.setFormatter(formatter)
            self.logger.addHandler(handler)
            self.logger.setLevel(logging.INFO)

        self.devices = dict()

    def new_device_found(self, device: FastbootDevice):  # pragma: nocover
        pass

    def device_removed(self, device: FastbootDevice):  # pragma: nocover
        pass

    def update_device_list(self):
        try:
            fbdevs_found = set()
            for device in usb.core.find(find_all=True):
                try:
                    fbdev = FastbootDevice(device)
                    fbdevs_found.add(fbdev)
                except ValueError:
                    # Ignore "The device is not a fastboot-compatible" errors
                    pass

            for device in set(self.devices.keys()) - {fbdev.device for fbdev in fbdevs_found}:
                self.logger.info(f"The fastboot device {repr(device)} got removed")

                try:
                    self.device_removed(self.devices[device])
                except Exception:
                    traceback.print_exc()

                del self.devices[device]

            for fbdev in fbdevs_found:
                if fbdev.device not in self.devices:
                    self.logger.info(f"Found a new fastboot device: {repr(device)}")
                    self.devices[fbdev.device] = fbdev

                    try:
                        self.new_device_found(fbdev)
                    except Exception:
                        traceback.print_exc()
        except Exception:
            traceback.print_exc()



from unittest.mock import MagicMock
import pytest
import copy
import time

from server.pdu import PDUState
from server.pdu.drivers.snmp import ManualSnmpPDU


@pytest.fixture(autouse=True)
def reset_easysnmp_mock(monkeypatch):
    import server.pdu.drivers.snmp as snmp

    global Session, time_sleep
    m1, m2 = MagicMock(), MagicMock()
    # REVIEW: I wonder if there's a clever way of covering the
    # difference in import locations between here and snmp.py
    monkeypatch.setattr(snmp, "Session", m1)
    monkeypatch.setattr(time, "sleep", m2)
    Session = m1
    time_sleep = m2


def test_driver_ManualSnmpPDU_check_OIDs_and_default_actions():
    pdu = ManualSnmpPDU("MyPDU", config={
        "hostname": "127.0.0.1",
        "system_id": "1.2.3.4",
        "outlet_labels": "5.6.7.8",
        "outlet_status": "5.6.7.9",
        "outlet_ctrl": "5.6.7.10",
        "state_mapping": {
            "on": 1,
            "off": 2,
            "reboot": 3,
        },
    })

    Session.assert_called_with(hostname=pdu.config['hostname'], community="private", version=1)

    assert pdu.outlet_labels == "5.6.7.8"
    assert pdu.outlet_status_oid(10) == "1.3.6.1.4.1.1.2.3.4.5.6.7.9.10"
    assert pdu.outlet_ctrl_oid(10) == "1.3.6.1.4.1.1.2.3.4.5.6.7.10.10"
    assert pdu.state_mapping.keys() == set([PDUState.ON, PDUState.OFF, PDUState.REBOOT])
    assert pdu.inverse_state_mapping.keys() == set([1, 2, 3])
    for k, _ in pdu.state_mapping.items():
        assert pdu.inverse_state_mapping[pdu.state_mapping[k]] == k


def test_driver_ManualSnmpPDU_check_v3_codepath():
    config = {
        "hostname": "127.0.0.1",
        "system_id": "1.2.3.4",
        "outlet_labels": "5.6.7.8",
        "outlet_status": "5.6.7.9",
        "outlet_ctrl": "5.6.7.10",
        "version": 3,
        "state_mapping": {
            "on": 1,
            "off": 2,
            "reboot": 3,
        },
    }

    # No username, authentication nor privacy
    pdu = ManualSnmpPDU("MyPDU", config=config)
    Session.assert_called_with(hostname=pdu.config['hostname'], version=3, security_level='no_auth_or_privacy')
    Session.reset_mock()

    # Check that adding the following fields does not impact the security_level parameter
    config['security_username'] = 'security_username'
    config['privacy_password'] = 'privacy_password'
    config['auth_password'] = 'auth_password'
    config['context_engine_id'] = 'context_engine_id'
    config['security_engine_id'] = 'security_engine_id'
    pdu = ManualSnmpPDU("MyPDU", config=config)
    Session.assert_called_with(hostname=pdu.config['hostname'], version=3, security_username='security_username',
                               privacy_password='privacy_password', auth_password='auth_password',
                               context_engine_id='context_engine_id', security_engine_id='security_engine_id',
                               security_level='no_auth_or_privacy')
    Session.reset_mock()

    # Check auth_without_privacy
    config['auth_protocol'] = 'auth_protocol'
    pdu = ManualSnmpPDU("MyPDU", config=config)
    Session.assert_called_with(hostname=pdu.config['hostname'], version=3, security_username='security_username',
                               privacy_password='privacy_password', auth_password='auth_password',
                               context_engine_id='context_engine_id', security_engine_id='security_engine_id',
                               auth_protocol='auth_protocol', security_level='auth_without_privacy')
    Session.reset_mock()

    # Check auth_with_privacy
    config['privacy_protocol'] = 'privacy_protocol'
    pdu = ManualSnmpPDU("MyPDU", config=config)
    Session.assert_called_with(hostname=pdu.config['hostname'], version=3, security_username='security_username',
                               privacy_password='privacy_password', auth_password='auth_password',
                               context_engine_id='context_engine_id', security_engine_id='security_engine_id',
                               auth_protocol='auth_protocol', privacy_protocol='privacy_protocol',
                               security_level='auth_with_privacy')
    Session.reset_mock()

    # Verify that we cannot have a privacy protocol without an auth protocol
    del config['auth_protocol']
    with pytest.raises(ValueError):
        pdu = ManualSnmpPDU("MyPDU", config=config)


def test_driver_ManualSnmpPDU_invalid_snmp_version():
    config = {
        "hostname": "127.0.0.1",
        "system_id": "1.2.3.4",
        "outlet_labels": "5.6.7.8",
        "outlet_status": "5.6.7.9",
        "outlet_ctrl": "5.6.7.10",
        "version": 3,
        "state_mapping": {
            "on": 1,
            "off": 2,
            "reboot": 3,
        },
    }

    ManualSnmpPDU("MyPDU", config=config)
    with pytest.raises(ValueError):
        config['version'] = 4
        ManualSnmpPDU("MyPDU", config=config)


def test_driver_ManualSnmpPDU_invalid_actions():
    with pytest.raises(ValueError):
        ManualSnmpPDU("MyPDU", config={
            "hostname": "127.0.0.1",
            "system_id": "1.2.3.4",
            "outlet_labels": "5.6.7.8",
            "outlet_status": "5.6.7.9",
            "outlet_ctrl": "5.6.7.10",
            "state_mapping": {
                "on": "FUDGE",
                "off": 2,
                "reboot": 3,
            },
        })


def test_driver_ManualSnmpPDU_missing_actions():
    with pytest.raises(AssertionError):
        ManualSnmpPDU("MyPDU", config={
            "hostname": "127.0.0.1",
            "system_id": "1.2.3.4",
            "outlet_labels": "5.6.7.8",
            "outlet_status": "5.6.7.9",
            "outlet_ctrl": "5.6.7.10",
            "state_mapping": {
                "off": 2,
                "reboot": 3,
            },
        })


def test_driver_ManualSnmpPDU_missing_parameters():
    valid_config = {
        "hostname": "127.0.0.1",
        "system_id": "1.2.3.4",
        "outlet_labels": "5.6.7.8",
        "outlet_status": "5.6.7.9",
        "outlet_ctrl": "5.6.7.10",
        "state_mapping": {
            "on": 1,
            "off": 2,
            "reboot": 3,
        },
    }

    ManualSnmpPDU("MyPDU", config=valid_config)
    for required_param in ["hostname", "system_id", "outlet_labels", "outlet_status", "state_mapping"]:
        new_config = copy.deepcopy(valid_config)
        del new_config[required_param]
        with pytest.raises((ValueError, AssertionError, KeyError)):
            ManualSnmpPDU("MyPDU", config=new_config)


def test_driver_ManualSnmpPDU_weird_inverses():
    valid_config = {
        "hostname": "127.0.0.1",
        "system_id": "1.2.3.4",
        "outlet_labels": "5.6.7.8",
        "outlet_status": "5.6.7.9",
        "outlet_ctrl": "5.6.7.10",
        "state_mapping": {
            "on": 1,
            "off": 2,
            "reboot": 3,
        },
        "inverse_state_mapping": {
            "on": 2,
            "off": 3,
            "reboot": 4,
        },
    }

    pdu = ManualSnmpPDU("MyPDU", config=valid_config)
    for k, _ in pdu.state_mapping.items():
        assert pdu.inverse_state_mapping[pdu.state_mapping[k]] == k + 1



from datetime import datetime
from enum import IntEnum

import time
from logging import getLogger, getLevelName, Formatter, StreamHandler


logger = getLogger(__name__)
logger.setLevel(getLevelName('INFO'))
log_formatter = \
    Formatter("%(asctime)s [%(threadName)s] [%(levelname)s] %(funcName)s: "
              "%(message)s")
console_handler = StreamHandler()
console_handler.setFormatter(log_formatter)
logger.addHandler(console_handler)


class PDUState(IntEnum):
    UNKNOWN = 0
    OFF = 1
    ON = 2
    REBOOT = 3

    @classmethod
    def valid_actions(cls):
        return [s for s in PDUState if s.value > 0]

    @property
    def is_valid_action(self):
        return self in self.valid_actions()


class PDUPort:
    def __init__(self, pdu, port_id, label=None, min_off_time=5):
        self.pdu = pdu
        self.port_id = port_id
        self.label = label
        self.min_off_time = min_off_time

        self.last_polled = None
        self.last_known_state = PDUState.UNKNOWN
        self.last_shutdown = None

    @property
    def reserved(self):
        return str(self.port_id) in self.pdu.reserved_port_ids

    def set(self, state):
        if self.reserved:
            logger.error("Port is reserved")
            return

        # Check the current state before writing it
        cur_state = self.state
        logger.debug("set: %s -> %s", cur_state.name, state.name)
        if cur_state == state:
            return

        if cur_state == PDUState.OFF and state == PDUState.ON:
            # Enforce a minimum amount of time between state changes
            time_spent_off = (datetime.now() - self.last_shutdown).total_seconds()
            if time_spent_off < self.min_off_time:
                time.sleep(self.min_off_time - time_spent_off)

        self.pdu.set_port_state(self.port_id, state)

        # Update the port's properties
        if state == PDUState.OFF and self.last_known_state != PDUState.OFF:
            self.last_shutdown = datetime.now()
        self.last_known_state = state

    @property
    def state(self):
        self.last_polled = datetime.now()

        state = self.pdu.get_port_state(self.port_id)

        # Reset the state
        if self.last_known_state != PDUState.OFF and state == PDUState.OFF:
            self.last_shutdown = self.last_polled

        self.last_known_state = state

        return state

    def __eq__(self, other):
        for attr in ["pdu", "port_id", "label", "min_off_time"]:
            if getattr(self, attr, None) != getattr(other, attr, None):
                return False
        return True

    def __repr__(self):
        return (f"<PDU Port: PDU={self.pdu.name}, ID={self.port_id}, label={self.label}, "
                f"min_off_time={self.min_off_time}>")


class PDU:
    # NOTICE: Please pass the config parameter directly from the user,
    # without modifications.
    def __init__(self, name, config, reserved_port_ids=[]):
        self.name = name
        self.config = config
        self.reserved_port_ids = set([str(p) for p in reserved_port_ids])

        # FIXME: It would be better to poll on state changes than wait
        # arbitrary amounts of time. Abstracting state transitions
        # into a method, so that subclasses can do more intelligent
        # things (in SNMP, a better thing can be done, for instance),
        # would be a start. There's a couple things in play for these pauses,
        #  1. The firmware on different PDUs needs differing amounts
        #  of time to settle in state transitions.  2. SNMP PDUs
        #  typically offer a way for the user to add state transition
        #  delays of their own, and these come with defaults that we
        #  need to wait *at least* that long for. We should do a
        #  better job of provisioning these values to suit our needs.
        if not hasattr(self, 'state_transition_delay_seconds'):
            self.state_transition_delay_seconds = 1

    @property
    def ports(self):
        # NOTICE: Left for drivers to implement
        # WARNING: Try to keep the list of returned ports constant, but do
        #          update the port properties (label / reservation) on every
        #          call. This allows keeping port metadata such as
        #          `last_shutdown` accurate across multiple polls.
        return []

    def set_port_state(self, port_id, state):
        # NOTICE: Left for drivers to implement
        return False

    def get_port_state(self, port_id):
        # NOTICE: Left for drivers to implement
        return PDUState.UNKNOWN

    def reserve_port(self, port_id):
        self.reserved_port_ids.add(str(port_id))

    def unreserve_port(self, port_id):
        try:
            self.reserved_port_ids.remove(str(port_id))
        except KeyError:
            pass

    @property
    def default_min_off_time(self):
        return 30

    def __eq__(self, other):
        return all([
            getattr(self, attr, None) == getattr(other, attr, None)
            for attr in ["name",
                         "config",
                         "reserved_port_ids"]])

    @classmethod
    def supported_pdus(cls):
        from .drivers.apc import ApcMasterswitchPDU
        from .drivers.cyberpower import PDU41004, PDU15SWHVIEC12ATNET
        from .drivers.dummy import DummyPDU
        from .drivers.virtual import VirtualPDU
        from .drivers.snmp import ManualSnmpPDU
        from .drivers.shelly import ShellyPDU
        from .drivers.tasmota import TasmotaPDU
        from .drivers.devantech import DevantechPDU
        from .drivers.kernelchip import KernelChipPDU
        from .drivers.usbhub import USBHubPDU

        return {
            "apc_masterswitch": ApcMasterswitchPDU,
            "cyberpower_pdu41004": PDU41004,
            "cyberpower_pdu15swhviec12atnet": PDU15SWHVIEC12ATNET,
            "dummy": DummyPDU,
            "vpdu": VirtualPDU,
            "snmp": ManualSnmpPDU,
            "shelly": ShellyPDU,
            "tasmota": TasmotaPDU,
            "devantech": DevantechPDU,
            "kernelchip": KernelChipPDU,
            "usbhub": USBHubPDU,
        }

    @classmethod
    def create(cls, model_name, pdu_name, config, reserved_port_ids=[]):
        driver = cls.supported_pdus().get(model_name)

        if driver is None:
            raise ValueError(f"Unknown model name '{model_name}'")

        return driver(pdu_name, config, reserved_port_ids)



from threading import Thread, Event
from dataclasses import asdict, field
from datetime import datetime, timedelta
from ipaddress import IPv4Address
from pathlib import Path
from pprint import pprint
from typing import Optional, Literal

from deepdiff import DeepDiff
from pydantic.dataclasses import dataclass
from pydantic import field_validator, PositiveInt
from inotify_simple import INotify, flags
import errno
import fcntl
import threading
import traceback
import time
import yaml
import os

from .dhcpd import MacAddress
from .logger import logger
from .dut import DUT
from . import config
from . import gitlab


@dataclass
class ConfigPDU:
    driver: str
    config: dict
    reserved_port_ids: list[str] = field(default_factory=list)


@dataclass
class ConfigGitlabRunner:
    token: str = "<invalid default>"
    exposed: bool = True
    runner_id: int = -1

    def verify_or_renew_token(self, gl, description, tags):
        def remove_and_register():
            logger.warning(f"{log_prefix}: Starting the renewal process...")

            if not self.remove(gl):
                logger.error(f"{log_prefix}: Could not unregister the runner on {gl.name}")
                return

            if gl.has_valid_looking_registration_token:
                runner = gitlab.register_runner(gitlab_url=gl.url,
                                                registration_token=gl.registration_token,
                                                description=description,
                                                tag_list=tags,
                                                maximum_timeout=gl.maximum_timeout,
                                                runner_type=gl.runner_type,
                                                group_id=gl.group_id,
                                                project_id=gl.project_id)
                if runner:
                    self.token = runner.token
                    self.runner_id = runner.id
                    logger.info(f"{log_prefix}: Got assigned a token")
                else:
                    logger.error(f"{log_prefix}: Could not register the runner on {gl.name}")
            else:
                logger.error(f"{log_prefix}: No registration tokens specified. Aborting...")

        log_prefix = f"{description}'s {gl.name} runner"

        logger.debug(f"{log_prefix}: Verifying the token")
        if not gitlab.verify_runner_token(gitlab_url=gl.url,
                                          token=self.token) or self.runner_id < 0:
            logger.warning(f"{log_prefix}: The token {self.token} is invalid.")
            remove_and_register()
        else:
            # The runner token is valid, let's check that the tags and description match!
            if gl.has_valid_looking_access_token:
                runner = gitlab.runner_details(gitlab_url=gl.url, private_token=gl.access_token,
                                               runner_id=self.runner_id)

                if runner is not None:
                    needs_re_registering = False
                    if runner.description != description:
                        logger.warning(f"{log_prefix}: The runner's description does not match the local value.")
                        needs_re_registering = True
                    elif set(runner.tag_list) != set(tags):
                        logger.warning(f"{log_prefix}: The runner tags list does not match the local database.")
                        needs_re_registering = True
                    elif runner.maximum_timeout != gl.maximum_timeout:
                        logger.warning(f"{log_prefix}: The runner's maximum timeout does not match the local value.")
                        needs_re_registering = True

                    if needs_re_registering:
                        remove_and_register()
            else:
                logger.warning(f"{log_prefix}: No access token specified, skipping tags verification")

    def remove(self, gl):
        if self.token == "<invalid default>":
            return True

        print(f"Unregister {gl.name}'s runner token")
        if not gitlab.unregister_runner(gitlab_url=gl.url, token=self.token):
            # We failed to remove the runner, make sure the token is still valid
            if gitlab.verify_runner_token(gitlab_url=gl.url, token=self.token):
                return False

        self.token = "<invalid default>"

        return True


@dataclass
class ConfigDUT:
    base_name: str
    ip_address: str  # TODO: Make sure all machines have a unique IP
    tags: list[str]
    manual_tags: list[str] = field(default_factory=list)
    local_tty_device: Optional[str] = None
    gitlab: dict[str, ConfigGitlabRunner] = field(default_factory=dict)
    pdu: Optional[str] = None
    pdu_port_id: Optional[str | int] = None
    pdu_off_delay: float = 30
    ready_for_service: bool = False
    is_retired: bool = False
    first_seen: datetime = field(default_factory=lambda: datetime.now())
    comment: Optional[str] = None

    mac_address: Optional[str] = None

    @field_validator('pdu_port_id')
    @classmethod
    def convert_pdu_port_id_to_str(cls, v):
        return str(v)

    @field_validator('mac_address')
    @classmethod
    def mac_address_is_understood_by_boots(cls, v):
        MacAddress(v)
        return v

    @field_validator('ip_address')
    @classmethod
    def ip_address_is_valid(cls, v):
        IPv4Address(v)
        return str(v)

    @property
    def full_name(self):
        # Get the index of the dut by looking at how many duts with
        # the same base name were registered *before* us.
        idx = 1
        for dut in self.mars_db.duts.values():
            if dut.base_name == self.base_name and dut.first_seen < self.first_seen:
                idx += 1

        return f"{config.FARM_NAME}-{self.base_name}-{idx}"

    # List of attributes that are safe to expose publicly
    @property
    def safe_attributes(self):
        return {
            "id": self.id,
            "base_name": self.base_name,
            "full_name": self.full_name,
            "tags": self.tags,
            "manual_tags": self.manual_tags,
            "mac_address": self.mac_address,
            "ip_address": self.ip_address,
            "local_tty_device": self.local_tty_device,
            "ready_for_service": self.ready_for_service,
            "comment": self.comment
        }

    @property
    def all_tags(self):
        farm_tags = [f"farm:{config.FARM_NAME}"]
        return sorted(set(self.tags + self.manual_tags + farm_tags))

    @property
    def available(self):
        return self.ready_for_service and not self.is_retired

    def expose_on_forges(self):
        # Make sure every gitlab instance is represented in the DUT's config
        for gl in self.mars_db.gitlab.values():
            if self.gitlab.get(gl.name) is None:
                self.gitlab[gl.name] = ConfigGitlabRunner()

        if self.available:
            for gl in self.mars_db.gitlab.values():
                local_cfg = self.gitlab.get(gl.name)
                if gl.expose_runners and local_cfg.exposed:
                    local_cfg.verify_or_renew_token(gl, description=self.full_name, tags=self.all_tags)

    def remove_from_forges(self):
        # Un-register every associated runner
        for gl_name, local_cfg in self.gitlab.items():
            if gl := self.mars_db.gitlab.get(gl_name):
                local_cfg.remove(gl)


@dataclass
class ConfigGitlab:
    url: str
    registration_token: Optional[str] = None
    runner_type: Optional[Literal["instance_type", "group_type", "project_type"]] = "instance_type"
    group_id: Optional[PositiveInt] = None
    project_id: Optional[PositiveInt] = None
    access_token: Optional[str] = None
    expose_runners: bool = True
    maximum_timeout: PositiveInt = 21600
    gateway_runner: Optional[ConfigGitlabRunner] = None

    # Function called once all the objects have been converted from dict
    # to their dataclass equivalent
    def __post_init__(self):
        if self.gateway_runner is not None:
            self.gateway_runner.mars_db = self

    @field_validator("url")
    @classmethod
    def url_is_valid(cls, v):
        assert v.startswith("https://")
        return v

    def has_valid_looking_token(self, token):
        return isinstance(token, str) and len(token) >= 8

    @property
    def has_valid_looking_registration_token(self):
        return self.has_valid_looking_token(self.registration_token)

    @property
    def has_valid_looking_access_token(self):
        return self.has_valid_looking_token(self.access_token)

    @property
    def should_expose_gateway_runner(self):
        if not self.expose_runners or self.gateway_runner is None:
            return False

        return self.gateway_runner.exposed


@dataclass
class MarsDB:
    pdus: dict[str, ConfigPDU] = field(default_factory=dict)
    duts: dict[str, ConfigDUT] = field(default_factory=dict)
    gitlab: dict[str, ConfigGitlab] = field(default_factory=dict)

    def reset_taint(self):
        self._disk_state = asdict(self)

    @property
    def diff_from_disk_state(self):
        return DeepDiff(self._disk_state, asdict(self), ignore_order=True)

    @property
    def is_tainted(self):
        return len(self.diff_from_disk_state) > 0

    # Function called once all the objects have been converted from dict
    # to their dataclass equivalent
    def __post_init__(self):

        # Always ensure that the VPDU config is in the mars db if the config
        # param is set. This protects against other things that might
        # externally modify (or generate) the mars db (e.g. vivian from the
        # ci-tron project.)
        if config.EXECUTOR_VPDU_ENDPOINT:
            vpdu_config = {"hostname": config.EXECUTOR_VPDU_ENDPOINT}
            if "VPDU" in self.pdus:
                self.pdus["VPDU"].driver = "vpdu"
                self.pdus["VPDU"].config = vpdu_config
            else:
                self.pdus["VPDU"] = ConfigPDU(driver="vpdu", config=vpdu_config)

        # Since we do not want to repeat ourselves in the config file, the name
        # of objects is set in the parent dict. However, it is quite useful for
        # objects to know their names and have access to the DB. This function
        # adds it back!
        for name, pdu in self.pdus.items():
            pdu.name = name
            pdu.mars_db = self

        for dut_id, dut in self.duts.items():
            dut.id = dut_id
            dut.mars_db = self

        for name, gitlab_instance in self.gitlab.items():
            gitlab_instance.name = name
            gitlab_instance.mars_db = self

        self.reset_taint()

        # Perform migrations after resetting the taint so that if the content
        # is modified by any migration, we will end up writing it back to disk

        # 2024/03/21 Migration: If the MAC Address is not set but the dut ID
        # looks like one, make it the mac address too.
        for dut_id, dut in self.duts.items():
            if dut.mac_address is None:
                try:
                    MacAddress(dut_id)
                    dut.mac_address = dut_id
                except Exception:
                    pass

    @classmethod
    def from_file(cls, file_path):
        with open(file_path, 'r') as f:
            cls.__dbLock(f)
            data = yaml.safe_load(f)
            return cls(**data if data else {})

    def save(self, file_path):
        with open(file_path, 'w') as f:
            self.__dbLock(f)
            yaml.dump(asdict(self), f, sort_keys=False)
            f.flush()
            os.fsync(f.fileno())
        self.reset_taint()

    # __dbLock returns as soon as an exclusive lock is acquired, else will
    # retry 5 times over 5 seconds, and raise a RuntimeError if unable to
    # acquire the lock after that period.
    @staticmethod
    def __dbLock(file):
        for tries in range(5):
            try:
                fcntl.flock(file.fileno(), fcntl.LOCK_EX | fcntl.LOCK_NB)
            except OSError as e:
                if e.errno in [errno.EACCES, errno.EAGAIN]:
                    time.sleep(1)
                    continue
                raise RuntimeError(f"unable to lock file: {file.filename}")

            # lock successfully acquired
            return

        raise RuntimeError(f"timed out trying to lock file: {file.filename}")


class MarsDBAccess:
    def __init__(self, mars):
        self._mars = mars
        self.entered_at = None

    def __enter__(self):
        self._mars._db_lock.acquire()
        self.entered_at = time.monotonic()
        return self._mars._db

    def __exit__(self, *args):
        self._mars.save_db_if_needed()
        self._mars._db_lock.release()

        held_time = (time.monotonic() - self.entered_at) * 1000.0
        if held_time > 100:
            logger.warning(f"MaRS DB lock held for an unusually long time: {held_time:.1f} ms")


class Mars(Thread):
    def __init__(self, pdu_daemon):
        super().__init__(name='MarsClient')

        self.pdu_daemon = pdu_daemon

        self._db = None
        self._db_lock = threading.RLock()  # Reentrant lock

        self._duts = {}
        self._discover_data = {}

        self.stop_event = Event()

    @property
    def discover_data(self):
        if self._discover_data:
            delta = time.monotonic() - self._discover_data.get('started_at')
            if delta >= self._discover_data.get('timeout'):
                self._discover_data = {}

        return self._discover_data

    @discover_data.setter
    def discover_data(self, value):
        self._discover_data = value

    def machine_discovered(self, dut_data, update_if_already_exists=False):
        machine = self.get_machine_by_id(dut_data.get('id'), raise_if_missing=False)

        if machine is None and self.discover_data:
            with self.db:
                # A new machine is about to be added through the discovery process

                pdu_port_id = self.discover_data.get('port_id')
                pdu_name = self.discover_data.get('pdu')

                machine = None
                if pdu_port := self.get_pdu_port_by_name(pdu_name, pdu_port_id):
                    dut_data["pdu_port_id"] = pdu_port_id
                    dut_data["pdu"] = pdu_name
                    dut_data["pdu_off_delay"] = pdu_port.pdu.default_min_off_time
                else:
                    logger.error(f"Discovery failed: The PDU port {pdu_name}/{pdu_port_id} doesn't exist in our DB")
                    # NOTE: We still want to create the DUT

                # We used the discovery data, so let's remove all of it
                self.discover_data = {}

                return self.add_or_update_machine(dut_data)
        elif machine is None:
            logger.warning("New machine found, despite no discovery process being in progress")

        if update_if_already_exists:
            return self.add_or_update_machine(dut_data)

    @property
    def db(self):
        return MarsDBAccess(self)

    @property
    def known_machines(self):
        with self._db_lock:
            return list(self._duts.values())

    def get_machine_by_id(self, machine_id, raise_if_missing=False):
        with self._db_lock:
            machine = self._duts.get(machine_id)
            if machine is None:
                # We could not find the machine by its actual ID, try finding it by full_name instead!
                for dut in self._duts.values():
                    if dut.full_name == machine_id:
                        machine = dut
                        break

            if machine is None and raise_if_missing:
                raise ValueError(f"Unknown machine ID '{machine_id}'")
            return machine

    def _machine_update_or_create(self, db_dut):
        with self._db_lock:
            dut = self._duts.get(db_dut.id)
            if dut is None:
                dut = DUT(mars=self, db_dut=db_dut)
                self._duts[dut.id] = dut
            else:
                dut.config_changed(db_dut=db_dut)

            self._db.duts[dut.id] = db_dut

            return dut

    def get_pdu_by_name(self, pdu_name, raise_if_missing=False):
        with self.db as mars_db:
            for name, pdu_cfg in mars_db.pdus.items():
                if name == pdu_name:
                    return self.pdu_daemon.get_or_create(pdu_cfg.driver, pdu_cfg.name,
                                                         pdu_cfg.config, pdu_cfg.reserved_port_ids,
                                                         update_if_existing=True)

        if raise_if_missing:
            raise ValueError(f'PDU "{pdu_name}" does not exist')

    def get_pdu_port_by_name(self, pdu_name, pdu_port_id, raise_if_missing=False, timeout=None):
        with self.db:
            pdu = self.get_pdu_by_name(pdu_name, raise_if_missing)

            if port := pdu.get_port_by_id(pdu_port_id, timeout=timeout):
                return port

        if raise_if_missing:
            raise ValueError(f'PDU "{pdu_name}" does not have a port ID named {pdu_port_id}')

    def save_db_if_needed(self):
        # TODO: Raise here if the lock is not currently held (no idea how to do that with an RLock)

        with self._db_lock:
            if self._db.is_tainted:
                print("Write-back the MarsDB to disk, after some local changes:")
                pprint(self._db.diff_from_disk_state, indent=2)
                print()

                self._db.save(config.MARS_DB_FILE)

    def add_or_update_machine(self, fields: dict):
        with self._db_lock:
            dut_id = fields.pop("id")

            if db_dut := self._db.duts.get(dut_id):
                cur_state = asdict(db_dut)
                db_dut = ConfigDUT(**(cur_state | fields))
            else:
                db_dut = ConfigDUT(**fields)

            # TODO: Try to find a way not to have to add these fields
            db_dut.id = dut_id
            db_dut.mars_db = self._db

            machine = self._machine_update_or_create(db_dut)
            self.save_db_if_needed()

        return machine

    def remove_machine(self, machine_id):
        with self._db_lock:
            machine = self._duts.pop(machine_id, None)
            if machine:
                # Stop thread
                machine.stop_machine()

                # Remove the associated DUT in MaRS DB
                self._db.duts.pop(machine_id, None)

                # Kill the gitlab runner token
                machine.remove_from_forges()

                # Save any change that may have happened after reloading
                self.save_db_if_needed()

                return True

            return False

    def sync_machines(self):
        with self.db:
            self._db = MarsDB.from_file(config.MARS_DB_FILE)

            local_only_machines = set(self.known_machines)
            for m in self._db.duts.values():

                machine = self._machine_update_or_create(m)

                # Remove the machine from the list of local-only machines
                local_only_machines.discard(machine)

            # Delete all the machines that are not found in MaRS
            for machine in local_only_machines:
                self._duts[machine.id].stop_machine()
                del self._duts[machine.id]

        # NOTE: Release the lock before checking doing any sort of IO

        # Expose the gateway runners
        for gl in self._db.gitlab.values():
            if gl.should_expose_gateway_runner:
                gl.gateway_runner.verify_or_renew_token(gl,
                                                        description=f"{config.FARM_NAME}-gateway",
                                                        tags=gitlab.generate_gateway_runner_tags())

        # Configure the DUTs
        for m in self._db.duts.values():
            # Expose the DUTs on all the forges
            m.expose_on_forges()

        # Update the gitlab runner configuration
        gitlab.generate_runner_config(self._db)

        # Save any change that may have happened after reloading
        self.save_db_if_needed()

    def stop(self, wait=True):
        self.stop_event.set()

        # Signal all the executors we want to stop
        for machine in self.known_machines:
            machine.stop_event.set()

        if wait:
            self.join()

    def join(self):
        for machine in self.known_machines:
            machine.join()
        super().join()

    def run(self):
        # Make sure the config file exists
        Path(config.MARS_DB_FILE).touch(exist_ok=True)

        # Set up a watch
        inotify = INotify()
        watch_flags = flags.CREATE | flags.DELETE | flags.MODIFY | flags.DELETE_SELF
        inotify.add_watch(config.MARS_DB_FILE, watch_flags)

        # Now wait for changes to the file
        last_sync = None
        while not self.stop_event.is_set():
            try:
                reason = None
                if last_sync is None:
                    reason = "Initial boot"
                elif len(inotify.read(timeout=1000)) > 0:
                    reason = "Got updated on disk"
                elif datetime.now() - last_sync > timedelta(minutes=30):
                    reason = "Periodic check"

                if reason:
                    logger.info(f"Syncing the MaRS DB. Reason: {reason}")

                    self.sync_machines()
                    last_sync = datetime.now()
            except Exception:
                traceback.print_exc()
                logger.info("Trying again in 60 seconds")
                time.sleep(60)



from unittest.mock import patch

import socket
import os

from server.socketactivation import (
    named_fds,
    get_sockets_by_name
)


def test_named_fds():
    # Make sure no FDs are found when LISTEN_PID is unset
    assert named_fds() == {}

    # Make sure no FDs are found when LISTEN_PID is wrong
    os.environ["LISTEN_PID"] = str(os.getpid() + 1)
    assert named_fds() == {}

    # Valid LISTEN_PID, but LISTEN_FDS is missing
    os.environ["LISTEN_PID"] = str(os.getpid())
    assert named_fds() == {}

    # Invalid LISTEN_FDS
    os.environ["LISTEN_FDS"] = "invalid"
    assert named_fds() == {}

    # Valid LISTEN_FDS, but no names
    os.environ["LISTEN_FDS"] = "4"
    assert named_fds() == {None: [3, 4, 5, 6]}

    # With some names missing
    os.environ["LISTEN_FDNAMES"] = "name1:name2:name1"
    assert named_fds() == {"name1": [3, 5], "name2": [4], None: [6]}

    # With too many names (last names ignored)
    os.environ["LISTEN_FDNAMES"] = "name1:name2:name1:name3:name4"
    assert named_fds() == {"name1": [3, 5], "name2": [4], "name3": [6]}


@patch("server.socketactivation.named_fds")
def test_get_sockets_by_name(named_fds):
    with (socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s1,
          socket.socket(socket.AF_INET6, socket.SOCK_DGRAM, 0) as s2,
          socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s3):

        named_fds.return_value = {"http": [s1.fileno(), s3.fileno()], None: [s2.fileno()]}

        # Check the default socket parameters
        http_sockets = get_sockets_by_name("http")
        assert len(http_sockets) == 2
        for s in http_sockets:
            assert s.family == socket.AF_INET
            assert s.type == socket.SOCK_STREAM
            assert s.proto == 0

        # Check no name, but specified
        noname_sockets = get_sockets_by_name(family=socket.AF_INET6, type=socket.SOCK_DGRAM, proto=0)
        assert len(noname_sockets) == 1
        for s in noname_sockets:
            assert s.family == socket.AF_INET6
            assert s.type == socket.SOCK_DGRAM
            assert s.proto == 0



from dataclasses import fields, field, asdict
from datetime import datetime, timedelta
from collections import defaultdict
from pydantic.dataclasses import dataclass
from pydantic.functional_validators import AfterValidator
from pydantic import field_validator, PositiveInt, NonNegativeInt, BaseModel, Field, HttpUrl
from typing import Annotated, Any, Dict, Generic, List, Union, Optional, TypeVar
from jinja2 import Template, ChainableUndefined
import yaml
import re

from . import config


@dataclass(config=dict(extra="forbid"))
class Target:
    id: Optional[str] = None
    tags: list[str] = None

    # Function called once all the objects have been converted from dict
    # to their dataclass equivalent and field validation succeeded
    def __post_init__(self):
        if not self.id and self.tags is None:
            raise ValueError("The target is neither identified by tags or id. "
                             "Use empty tags to mean 'any machines'.")

        if self.tags is None:
            self.tags = []

    def __str__(self):
        return f"<Target: id={self.id}, tags={self.tags}>"


@dataclass(config=dict(extra="forbid"))
class Timeout:
    days: Optional[float] = None
    hours: Optional[float] = None
    minutes: Optional[float] = None
    seconds: Optional[float] = None
    milliseconds: Optional[float] = None
    retries: NonNegativeInt = 0

    @classmethod
    def create(cls, name, *args, **kwargs):
        timeout = cls(*args, **kwargs)
        timeout.name = name
        return timeout

    # Function called once all the objects have been converted from dict
    # to their dataclass equivalent and field validation succeeded
    def __post_init__(self):
        days = self.days or 0
        hours = self.hours or 0
        minutes = self.minutes or 0
        seconds = self.seconds or 0
        milliseconds = self.milliseconds or 0

        self.timeout = timedelta(days=days, hours=hours,
                                 minutes=minutes, seconds=seconds,
                                 milliseconds=milliseconds)

        if (self.days is None and self.hours is None and self.minutes is None and
                self.seconds is None and self.milliseconds is None):
            self.timeout = timedelta.max

        self.started_at = None
        self.retried = 0

    @property
    def active_for(self):
        if self.started_at is not None:
            return datetime.now() - self.started_at
        else:
            return None

    @property
    def is_started(self):
        return self.started_at is not None

    @property
    def has_expired(self):
        active_for = self.active_for
        return active_for is not None and active_for > self.timeout

    def start(self):
        self.started_at = datetime.now()

    def reset(self, when=None):
        if when is None:
            when = datetime.now()
        self.started_at = when

    def retry(self):
        self.stop()
        self.retried += 1

        return self.retried <= self.retries

    def stop(self):
        self.started_at = None

    def __str__(self):
        return f"<Timeout {self.name}: value={self.timeout}, retries={self.retried}/{self.retries}>"


@dataclass(config=dict(extra="forbid"))
class Timeouts:
    overall: Timeout = field(default_factory=lambda: Timeout(hours=6))
    infra_setup: Timeout = field(default_factory=Timeout)
    infra_teardown: Timeout = field(default_factory=lambda: Timeout(minutes=10))
    boot_cycle: Timeout = field(default_factory=Timeout)
    console_activity: Timeout = field(default_factory=Timeout)
    first_console_activity: Timeout = field(default_factory=Timeout)
    watchdogs: dict[str, Timeout] = field(default_factory=dict)

    # Function called once all the objects have been converted from dict
    # to their dataclass equivalent and field validation succeeded
    def __post_init__(self):
        # Make sure to add the name to every field
        for f in fields(self):
            if f.type is Timeout:
                getattr(self, f.name).name = f.name

        # Add the watchdogs' names
        for name, wd in self.watchdogs.items():
            wd.name = name

        # Ensure that the overall and tear-down timeouts have retries=0
        for timeout in [self.overall, self.infra_teardown]:
            if timeout.retries != 0:
                raise ValueError("Neither the overall nor the teardown timeout can have retries")

    def __iter__(self):
        for f in fields(self):
            if f.type is Timeout:
                yield getattr(self, f.name)

        for wd in self.watchdogs.values():
            yield wd

    @property
    def expired_list(self):
        expired = []
        for timeout in self:
            if timeout.has_expired:
                expired.append(timeout)
        return expired

    @property
    def has_expired(self):
        return len(self.expired_list) > 0


@dataclass(config=dict(extra="forbid"))
class Pattern:
    regex: str

    @field_validator("regex")
    @classmethod
    def convert_to_regex(cls, v):
        try:
            return re.compile(v.encode())
        except re.error as e:
            raise ValueError(f"Console pattern '{v}' is not a valid regular expression: {e.msg}")

    def __str__(self):
        return f"{self.regex.pattern}"


@dataclass(config=dict(extra="forbid"))
class Watchdog:
    start: Pattern
    reset: Pattern
    stop: Pattern

    # Function called once all the objects have been converted from dict
    # to their dataclass equivalent and field validation succeeded
    def __post_init__(self):
        self.timeout = None

    def set_timeout(self, timeout):
        self.timeout = timeout

    def process_line(self, line):
        # Do not parse lines if no timeout is associated
        if self.timeout is None:
            return {}

        if not self.timeout.is_started:
            if self.start.regex.search(line):
                self.timeout.start()
                return {"start"}
        else:
            if self.reset.regex.search(line):
                self.timeout.reset()
                return {"reset"}
            elif self.stop.regex.search(line):
                self.timeout.stop()
                return {"stop"}

        return {}

    # I would have loved to re-use `stop()` here, but it collides with the stop pattern
    def cancel(self):
        if self.timeout is not None:
            self.timeout.stop()


@dataclass(config=dict(extra="forbid"))
class ConsoleState:
    session_end: Pattern = field(default_factory=lambda: Pattern(regex=r"^\[[\d \.]{12}\] reboot: Power Down$"))
    session_reboot: Optional[Pattern] = None
    job_success: Optional[Pattern] = None
    job_warn: Optional[Pattern] = None
    machine_unfit_for_service: Optional[Pattern] = None
    watchdogs: dict[str, Watchdog] = field(default_factory=dict)

    # Function called once all the objects have been converted from dict
    # to their dataclass equivalent and field validation succeeded
    def __post_init__(self):
        self._patterns = dict()
        self._matched = set()

        if self.machine_unfit_for_service is None and config.CONSOLE_PATTERN_DEFAULT_MACHINE_UNFIT_FOR_SERVICE_REGEX:
            self.machine_unfit_for_service = Pattern(config.CONSOLE_PATTERN_DEFAULT_MACHINE_UNFIT_FOR_SERVICE_REGEX)

        # Generate the list of patterns to match
        for f in fields(self):
            if f.type in [Pattern, Optional[Pattern]]:
                pattern = getattr(self, f.name, None)
                if pattern:
                    pattern.name = f.name
                    self._patterns[f.name] = pattern

    def process_line(self, line):
        # Try matching all the patterns
        matched = set()
        for name, pattern in self._patterns.items():
            if pattern.regex.search(line):
                matched.add(name)
        self._matched.update(matched)

        # Try matching the watchdogs
        for name, wd in self.watchdogs.items():
            _matched = wd.process_line(line)
            matched.update({f"{name}.{m}" for m in _matched})

        return matched

    def reset_per_boot_state(self):
        self._matched.discard("session_reboot")

    @property
    def session_has_ended(self):
        return "session_end" in self._matched or "unfit_for_service" in self._matched

    @property
    def needs_reboot(self):
        return "session_reboot" in self._matched

    @property
    def machine_is_unfit_for_service(self):
        return "machine_unfit_for_service" in self._matched

    @property
    def job_status(self):
        if "session_end" not in self._matched:
            return "INCOMPLETE"

        if "job_success" in self._patterns:
            if "job_success" in self._matched:
                if "job_warn" in self._matched:
                    return "WARN"
                else:
                    return "PASS"
            else:
                return "FAIL"
        else:
            return "COMPLETE"


class DeploymentMixin:
    def update(self, d):
        # Nothing to do if `d` is empty
        if not d:
            return

        # Convert `d` from dict to the proper instance if needed
        if type(d) is dict:
            d = type(self)(**d)

        # Assert that both `self` and `d` share the same type
        assert type(d) is type(self)

        # Check every field
        for f in fields(self):
            if new := getattr(d, f.name, None):
                # If the value was already set and the new value is a DeploymentMixin, call `update()` rather than
                # simply copying
                if cur := getattr(self, f.name):
                    if isinstance(cur, DeploymentMixin):
                        cur.update(new)
                        continue

                setattr(self, f.name, new)

    @classmethod
    def _add_artifacts_from_object(cls, artifacts, obj_name, obj):
        for url, paths in obj.artifacts.items():
            for path, artifact in paths.items():
                artifacts[url][(obj_name, ) + path] = artifact

    @property
    def artifacts(self):
        artifacts = defaultdict(dict)

        for f in fields(self):
            if f_value := getattr(self, f.name, None):
                if isinstance(f_value, DeploymentMixin):
                    self._add_artifacts_from_object(artifacts, f.name, f_value)
                elif f.name == "url":
                    artifacts[str(self.url)][()] = self

        return artifacts


T = TypeVar("T")


SingleItemOrList = T | List[T]


def is_valid_category_name(v: str, allow_keywords=False) -> str:
    if v[0] == ":":
        if not allow_keywords or v not in [":uncategorised"]:
            raise ValueError("User-defined category names cannot start with ':'")
    return v


Category = Annotated[str, AfterValidator(lambda v: is_valid_category_name(v, allow_keywords=False))]


CategoryOrKeyword = Annotated[str, AfterValidator(lambda v: is_valid_category_name(v, allow_keywords=True))]


JobCollectionOfLists = SingleItemOrList | Dict[CategoryOrKeyword, SingleItemOrList]


@dataclass(config=dict(extra="forbid"))
class CollectionOfLists(Generic[T], DeploymentMixin):
    categories: Optional[Dict[Category, List[T]]] = field(default_factory=dict)
    uncategorised: Optional[List[T]] = field(default_factory=list)

    @property
    def as_list(self) -> List[T]:
        # NOTE: We are sorting the categories by names to let users order how elements are staged
        cl = []
        for k in sorted(self.categories.keys()):
            cl.extend(self.categories[k])
        return cl + self.uncategorised

    def update(self, d: "CollectionOfLists"):
        assert isinstance(d, CollectionOfLists)

        self.categories.update(d.categories)

        if len(d.uncategorised) > 0:
            self.uncategorised = d.uncategorised

    def __str__(self):
        return " ".join([str(v) for v in self.as_list])

    def __eq__(self, other: "CollectionOfLists"):
        if isinstance(other, str):
            return str(self) == other
        else:
            for f in fields(self):
                if getattr(self, f.name, None) != getattr(other, f.name, None):
                    return False

            return True

    def __iter__(self):
        return iter(self.as_list)

    @classmethod
    def from_job(cls, v: "CollectionOfLists" | JobCollectionOfLists):
        def to_list(value: SingleItemOrList):
            if isinstance(value, list):
                return value
            else:
                return [value]

        categories = dict()
        uncategorised = []
        if isinstance(v, cls):
            return v
        elif isinstance(v, dict):
            uncategorised = v.pop(":uncategorised", [])
            for key, value in v.items():
                categories[key] = to_list(value)
        else:
            uncategorised = to_list(v)

        return cls(categories=categories, uncategorised=uncategorised)


ComplexList = Union[Annotated[JobCollectionOfLists, AfterValidator(lambda v: CollectionOfLists.from_job(v))],
                    CollectionOfLists[T]]


@dataclass(config=dict(extra="forbid"))
class FileHttpUrl(DeploymentMixin):
    url: Optional[HttpUrl] = None


@dataclass(config=dict(extra="forbid"))
class KernelDeployment(FileHttpUrl, DeploymentMixin):
    cmdline: ComplexList[str] = None


@dataclass(config=dict(extra="forbid"))
class StorageHttpArtifactDeployment(DeploymentMixin):
    path: str
    data: Optional[str | bytes] = None

    # Proposals:
    # args: Optional[Dict[str, str]] = None  # Only match when the following GET params are set


@dataclass(config=dict(extra="forbid"))
class StorageDeployment(DeploymentMixin):
    http: Optional[ComplexList[StorageHttpArtifactDeployment]] = field(default_factory=list)


@dataclass(config=dict(extra="forbid"))
class FastbootDeployment(DeploymentMixin):
    header_version: Optional[int] = None
    base: Optional[int] = None
    kernel_offset: Optional[int] = None
    ramdisk_offset: Optional[int] = None
    dtb_offset: Optional[int] = None
    tags_offset: Optional[int] = None
    board: Optional[str] = None
    pagesize: Optional[int] = None

    @property
    def fields_set(self):
        return {k: v for k, v in asdict(self).items() if v}

    def __str__(self):
        non_empty_fields = list()
        for key, value in self.fields_set.items():
            # Show some of the fields as hex values, since it is likely what people used
            if key in ["base"] or "_offset" in key:
                value = hex(value)
            non_empty_fields.append(f"{key}={value}")

        s = ", ".join(non_empty_fields)
        return f"<Fastboot: {s}>"


@dataclass(config=dict(extra="forbid"))
class DeploymentState(DeploymentMixin):
    kernel: Optional[KernelDeployment] = None
    initramfs: Optional[FileHttpUrl] = None
    dtb: Optional[FileHttpUrl] = None
    storage: Optional[StorageDeployment] = None
    fastboot: Optional[FastbootDeployment] = None

    @classmethod
    def object_url(cls, obj):
        if obj and obj.url:
            return str(obj.url)

    # NOTE: For backwards compatibility
    @property
    def kernel_url(self):
        return self.object_url(self.kernel)

    # NOTE: For backwards compatibility
    @property
    def kernel_cmdline(self):
        return str(self.kernel.cmdline) if self.kernel else None

    # NOTE: For backwards compatibility
    @property
    def initramfs_url(self):
        return self.object_url(self.initramfs)

    # NOTE: For backwards compatibility
    @property
    def dtb_url(self):
        return self.object_url(self.dtb)

    def __str__(self):
        str_fields = [
            f"kernel_url: {self.kernel_url}",
            f"initramfs_url: {self.initramfs_url}",
            f"kernel_cmdline: {self.kernel_cmdline}"
        ]

        # Do not print these parameters if their value is None
        for optional_field in ["dtb_url", "fastboot"]:
            if f_value := getattr(self, optional_field, None):
                str_fields.append(f"{optional_field}: {f_value}")

        str_fields = "\n    ".join(str_fields)
        return f"""<Deployment:
    {str_fields}>
"""


# NOTE: Because the "continue" field in deployment cannot be used as a dataclass variable,
# we have to open code it using a pydantic BaseModel. Fortunately, pydantic is happy to
# allow us to use an alias for the field.
class Deployment(BaseModel):
    model_config = dict(populate_by_name=True, extra="forbid")

    start: DeploymentState
    continue_: DeploymentState = Field(default=None, alias='continue')

    def __init__(self, /, **data: Any):
        super().__init__(**data)

        continue_ = DeploymentState(**asdict(self.start))
        if self.continue_:
            continue_.update(self.continue_)
        self.continue_ = continue_

    @property
    def artifacts(self):
        artifacts = defaultdict(dict)

        DeploymentMixin._add_artifacts_from_object(artifacts, "start", self.start)
        DeploymentMixin._add_artifacts_from_object(artifacts, "continue", self.continue_)

        return artifacts


@dataclass(config=dict(extra="forbid"))
class Job:
    console_patterns: ConsoleState
    deployment: Deployment

    version: PositiveInt = 1
    deadline: datetime = datetime.max
    target: Target = field(default_factory=Target)
    timeouts: Timeouts = field(default_factory=Timeouts)

    # Function called once all the objects have been converted from dict
    # to their dataclass equivalent and field validation succeeded
    def __post_init__(self):
        # Associate all the timeouts to their respective watchdogs
        for name, wd in self.console_patterns.watchdogs.items():
            wd.set_timeout(self.timeouts.watchdogs.get(name))

    # NOTE: For backwards compatibility
    @property
    def deployment_start(self):
        return self.deployment.start if self.deployment else None

    # NOTE: For backwards compatibility
    @property
    def deployment_continue(self):
        return self.deployment.continue_ if self.deployment else None

    @classmethod
    def render_with_resources(cls, job_str, machine=None, bucket=None, **kwargs):
        template_params = {
            "ready_for_service": machine.ready_for_service if machine else True,
            "machine_id": machine.id if machine else "machine_id",
            "machine": machine.safe_attributes if machine else {},
            "machine_tags": machine.tags if machine else [],
            "local_tty_device": machine.local_tty_device if machine else "",
            **{k.lower(): v for k, v in config.job_environment_vars().items()},
            **kwargs,
        }

        if bucket:
            dut_creds = bucket.credentials('dut')

            template_params["minio_url"] = bucket.minio.url
            template_params["job_bucket"] = bucket.name
            template_params["job_bucket_access_key"] = dut_creds.username
            template_params["job_bucket_secret_key"] = dut_creds.password

        rendered_job_str = Template(job_str, undefined=ChainableUndefined).render(**template_params)
        return cls(**yaml.safe_load(rendered_job_str))

    @classmethod
    def from_path(cls, job_template_path, machine=None, bucket=None):
        with open(job_template_path, "r") as f_template:
            template_str = f_template.read()
            return Job.render_with_resources(template_str, machine, bucket)

    def __str__(self):
        if len(self.timeouts.watchdogs) == 0:
            timeout_watchdogs = "None"
        else:
            timeout_watchdogs = ""
            for name, wd in self.timeouts.watchdogs.items():
                timeout_watchdogs += f"\n            {name}: {wd}"

        if len(self.console_patterns.watchdogs) == 0:
            patterns_watchdogs = "None"
        else:
            patterns_watchdogs = ""
            for name, wd in self.console_patterns.watchdogs.items():
                patterns_watchdogs += f"\n            {name}:"
                patterns_watchdogs += f"\n                start: {wd.start}"
                patterns_watchdogs += f"\n                reset: {wd.reset}"
                patterns_watchdogs += f"\n                stop:  {wd.stop}"

        return f"""<Job:
    version: {self.version}

    deadline: {self.deadline}
    target: {self.target}

    timeouts:
        overall:                {self.timeouts.overall}
        infra_setup:            {self.timeouts.infra_setup}
        boot_cycle:             {self.timeouts.boot_cycle}
        console_activity:       {self.timeouts.console_activity}
        first_console_activity: {self.timeouts.first_console_activity}
        watchdogs:              {timeout_watchdogs}

    console patterns:
        session_end:               {self.console_patterns.session_end}
        session_reboot:            {self.console_patterns.session_reboot}
        job_success:               {self.console_patterns.job_success}
        job_warn:                  {self.console_patterns.job_warn}
        machine_unfit_for_service: {self.console_patterns.machine_unfit_for_service}
        watchdogs:              {patterns_watchdogs}

    start deployment:
        kernel_url:     {self.deployment_start.kernel_url}
        initramfs_url:  {self.deployment_start.initramfs_url}
        kernel_cmdline: {self.deployment_start.kernel_cmdline}

    continue deployment:
        kernel_url:     {self.deployment_continue.kernel_url}
        initramfs_url:  {self.deployment_continue.initramfs_url}
        kernel_cmdline: {self.deployment_continue.kernel_cmdline}>"""



from .. import PDU, PDUPort, PDUState

import itertools
import socket
import time
from contextlib import contextmanager

# The driver for KernelChip Lauren-2, Laurent-112, Laurent-128 relays,
# controlled over IP network.
#
# Example protocol ('>' and '<' mark sent and received strings and are not a
# part of the protocol)
#
# < #FLG,AB,11,11
# < JConfig from FLASH
# > $KE
# < #OK
# > $KE,INF
# < #INF,Laurent-128,LX11,Y18J-K78C-2D51-M199
# > $KE,RDR,1
# < #Access denied. Password is needed.
# > $KE,PSW,SET,Laurent
# < #PSW,SET,OK
# > $KE,RDR,1
# < #RDR,1,0
# > $KE,REL,1,1
# < #REL,OK
# > $KE,RDR,1
# < #RDR,1,1
# > $KE,REL,1,0
# < #REL,OK
# > $KE,RDR,ALL
# < #RDR,ALL,0000000000000000000000000000


class KernelChipConnection:
    def __init__(self, socket):
        self.socket = socket
        self.reader = self.reader()

    def send_msg(self, *args):
        out = b','.join(itertools.chain([b'$KE'], args)) + b'\r\n'
        self.socket.sendall(out)

    def reader(self):
        chunks = []
        seen_cr = False
        while True:
            chunk = self.socket.recv(2048)
            pos = 0
            for (i, c) in enumerate(chunk):
                if c == ord('\r'):
                    if seen_cr:
                        raise ValueError("Duplicate '\\r'")
                    if i != pos:
                        chunks.append(chunk[pos:i])
                    pos = i + 1
                    seen_cr = True
                elif c == ord('\n'):
                    if seen_cr:
                        yield b''.join(chunks)
                        chunks = []
                        pos = i + 1
                        seen_cr = False
                else:
                    seen_cr = False
            if pos != len(chunk):
                chunks.append(chunk[pos:])

    def recv_msg(self):
        buf = next(self.reader)

        # Workarounds for the rogue message after boot
        if buf.startswith(b'#FLG,'):
            buf = next(self.reader)
        if buf.startswith(b'JConfig'):
            buf = next(self.reader)

        if buf[0] != ord('#'):
            raise ValueError('Invalid first byte')
        return buf[1:].split(b',')


# in seconds
PORT_STATE_TIMEOUT = 0.1


class KernelChipPDU(PDU):
    def __init__(self, name, config, reserved_port_ids=[]):
        self.host, self.port = config.get('hostname', 'localhost:2424').split(':')
        self.port = int(self.port)
        self.password = config.get('password', None)

        # Check for supported module
        with self.conn() as s:
            # Protocol check to make sure we speak to a compatible PDU.
            s.send_msg()
            r = s.recv_msg()[0]
            if r != b'OK':
                raise ValueError("Protocol error")

            self._auth(s)

            self._update_state(s)
            self.num_relay = len(self.cached_state)

        self._ports = [PDUPort(self, i) for i in range(1, self.num_relay + 1)]

        super().__init__(name, config, reserved_port_ids)

    def _auth(self, s):
        if self.password:
            s.send_msg(b'PSW', b'SET', bytes(self.password, 'ascii'))
            if s.recv_msg() != [b'PSW', b'SET', b'OK']:
                raise ValueError("Password Verification failure")

    def _is_cached_state_valid(self):
        return self.state_stamp and time.time() - self.state_stamp < PORT_STATE_TIMEOUT

    def _update_state(self, s):
        cur_time = time.time()

        s.send_msg(b'RDR', b'ALL')
        r = s.recv_msg()
        if r[0].startswith(b'Access'):
            raise ValueError("Password is required")

        if len(r) != 3 or r[0] != b'RDR' or r[1] != b'ALL':
            raise ValueError("Failure getting relays state")

        self.state_stamp = cur_time
        self.cached_state = r[2]

    @property
    def ports(self):
        return self._ports

    @contextmanager
    def conn(self, *args, **kw):
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            s.connect((self.host, self.port))

            yield KernelChipConnection(s)

    def set_port_state(self, port_id, state):
        if port_id < 1 or port_id > self.num_relay:
            raise ValueError('Invalid port id')

        port = bytes(str(port_id), 'ascii')
        msg = b'0' if state == PDUState.OFF else b'1'

        with self.conn() as s:
            self._auth(s)

            s.send_msg(b'REL', port, msg)
            r = s.recv_msg()

            # safety net, __init__ checks whether the password is required
            if r[0].startswith(b'Access'):  # pragma: no cover
                raise ValueError("Password is required")

            if r != [b'REL', b'OK']:
                raise ValueError("Incorrect response")

            self._update_state(s)

    def get_port_state(self, port_id):
        if port_id < 1 or port_id > self.num_relay:
            raise ValueError('Invalid port id')

        if not self._is_cached_state_valid():
            with self.conn() as s:
                self._auth(s)
                self._update_state(s)

        if self.cached_state[port_id - 1] == ord(b'1'):
            return PDUState.ON
        else:
            return PDUState.OFF



from .. import PDU, PDUPort, PDUState

import socket
from contextlib import contextmanager


class DevantechPDU(PDU):
    def __init__(self, name, config, reserved_port_ids=[]):
        self.host, self.port = config.get('hostname', 'localhost:17494').split(':')
        self.port = int(self.port)
        self.password = config.get('password', None)

        # mapping from module id -> number of relays
        self.devices = {18: 2,    # eth002
                        19: 8,    # eth008
                        20: 4,    # eth484
                        21: 20,   # eth8020
                        22: 4,    # wifi484
                        24: 20,   # wifi8020
                        26: 2,    # wifi002
                        28: 8,    # wifi008
                        51: 20,   # eth1620
                        52: 10}   # eth1610

        # Check for supported module
        with self.conn() as s:
            s.sendall(b"\x10")
            id = s.recv(3)[0]
            if id not in self.devices:
                raise ValueError("not supported module id found")

        self._ports = [PDUPort(self, i) for i in range(self.devices[id])]

        super().__init__(name, config, reserved_port_ids)

    @property
    def ports(self):
        return self._ports

    @contextmanager
    def conn(self, *args, **kw):
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

            s.connect((self.host, self.port))

            # Check to see if password is enabled
            s.sendall(b"\x7a")
            if s.recv(1)[0] == 0:
                passwordString = b'\x79' + self.password.encode()
                s.sendall(passwordString)
                if s.recv(1)[0] != 1:
                    raise ValueError("wrong password")

            yield s

    def set_port_state(self, port_id, state):
        msg = b'\x21' if state == PDUState.OFF else b'\x20'
        msg += int(port_id).to_bytes(1, 'big')
        msg += b'\x00'

        with self.conn() as s:
            s.sendall(msg)
            if s.recv(1)[0] == 1:
                raise ValueError('failed to set port state')

    def get_port_state(self, port_id):
        num_bytes = (len(self._ports) + 7) // 8

        with self.conn() as s:
            s.sendall(b'\x24')
            port_states = b''
            while len(port_states) != num_bytes:
                buf = s.recv(num_bytes - len(port_states), socket.MSG_WAITALL)
                if len(buf) == 0:
                    raise ValueError("we never received the expected amount of bytes")
                port_states += buf

            byte_index = port_id // 8
            bit_position = port_id % 8

            if port_states[byte_index] & (1 << bit_position):
                return PDUState.ON
            else:
                return PDUState.OFF



from unittest.mock import MagicMock, patch, mock_open
from dataclasses import dataclass

from server.gitlab import GitlabRunnerRegistration, register_runner, unregister_runner, runner_details
from server.gitlab import verify_runner_token, generate_runner_config, SanitizedFieldsMixin, Runner
from server.gitlab import generate_gateway_runner_tags
from server.mars import MarsDB
import server.config as config

from jinja2 import Template
import pytest


def test_SanitizedFieldsMixin__from_api():
    @dataclass
    class Dataclass(SanitizedFieldsMixin):
        field1: int
        field2: int

    assert Dataclass.from_api({"field1": 1, "field2": 2, "field3": 3}) == Dataclass(field1=1, field2=2)


def test_register_runner():
    url = "my url"
    registration_token = "reg_token"
    description = "toto"
    tag_list = ["tag1", "tag2"]

    runner_id = 1234
    runner_token = "my token"

    post_mock_return_value = MagicMock(status_code=201,
                                       json=MagicMock(return_value={"id": runner_id, "token": runner_token}))

    # Check the defaults
    with patch("server.gitlab.requests.post", return_value=post_mock_return_value) as post_mock:
        r = register_runner(gitlab_url=url, registration_token=registration_token,
                            description=description, tag_list=tag_list)
        assert r == GitlabRunnerRegistration(id=runner_id, token=runner_token)

        post_mock.assert_called_with(f"{url}/api/v4/user/runners",
                                     headers={"PRIVATE-TOKEN": registration_token},
                                     params={'description': description,
                                             'tag_list': ",".join(tag_list),
                                             'run_untagged': False,
                                             'maximum_timeout': 3600,
                                             'runner_type': 'instance_type'})

    # Check registration of a group runner
    with patch("server.gitlab.requests.post", return_value=post_mock_return_value) as post_mock:
        r = register_runner(gitlab_url=url, registration_token=registration_token,
                            description=description, tag_list=tag_list,
                            runner_type="group_type", group_id=1234)
        assert r == GitlabRunnerRegistration(id=runner_id, token=runner_token)

        post_mock.assert_called_with(f"{url}/api/v4/user/runners",
                                     headers={"PRIVATE-TOKEN": registration_token},
                                     params={'description': description,
                                             'tag_list': ",".join(tag_list),
                                             'run_untagged': False,
                                             'maximum_timeout': 3600,
                                             'runner_type': 'group_type',
                                             'group_id': 1234})

    # Check registration of a project runner
    with patch("server.gitlab.requests.post", return_value=post_mock_return_value) as post_mock:
        r = register_runner(gitlab_url=url, registration_token=registration_token,
                            description=description, tag_list=tag_list,
                            runner_type="project_type", project_id=1234)
        assert r == GitlabRunnerRegistration(id=runner_id, token=runner_token)

        post_mock.assert_called_with(f"{url}/api/v4/user/runners",
                                     headers={"PRIVATE-TOKEN": registration_token},
                                     params={'description': description,
                                             'tag_list': ",".join(tag_list),
                                             'run_untagged': False,
                                             'maximum_timeout': 3600,
                                             'runner_type': 'project_type',
                                             'project_id': 1234})

    # Check that invalid types are rejected
    with pytest.raises(AssertionError):
        register_runner(gitlab_url=url, registration_token=registration_token,
                        description=description, tag_list=tag_list,
                        runner_type="invalid")

    # Check that we don't allow setting a runner type without the accompanying parameter
    for runner_type in ["group_type", "project_type"]:
        with pytest.raises(AssertionError):
            register_runner(gitlab_url=url, registration_token=registration_token,
                            description=description, tag_list=tag_list,
                            runner_type=runner_type)

    # Check that errors during registration return None
    with patch("server.gitlab.requests.post", return_value=MagicMock(status_code=403)) as post_mock:
        r = register_runner(gitlab_url=url, registration_token=registration_token,
                            description=description, tag_list=tag_list)
        assert r is None


def test_unregister_runner():
    url = "my url"
    runner_token = "my token"

    with patch("server.gitlab.requests.delete", return_value=MagicMock(status_code=204)) as delete_mock:
        assert unregister_runner(gitlab_url=url, token=runner_token)
        delete_mock.assert_called_with(f"{url}/api/v4/runners", params={"token": runner_token})

    with patch("server.gitlab.requests.delete", return_value=MagicMock(status_code=200)) as delete_mock:
        assert not unregister_runner(gitlab_url=url, token=runner_token)


def test_verify_runner_token():
    url = "my url"
    runner_token = "my token"

    with patch("server.gitlab.requests.post", return_value=MagicMock(status_code=200)) as post_mock:
        assert verify_runner_token(gitlab_url=url, token=runner_token)
        post_mock.assert_called_with(f"{url}/api/v4/runners/verify", params={"token": runner_token})

    with patch("server.gitlab.requests.post", return_value=MagicMock(status_code=403)) as post_mock:
        assert not verify_runner_token(gitlab_url=url, token=runner_token)


def test_runner_details():
    url = "my url"
    private_token = "my token"
    runner_id = 42

    details = {
        "id": runner_id,
        "description": "keywords-gfx10-1",
        "ip_address": "local_ip",
        "active": True,
        "paused": False,
        "is_shared": True,
        "runner_type": "instance_type",
        "name": None,
        "online": True,
        "status": "online",
        "tag_list": [
            "TAG1",
            "TAG2"
        ],
        "run_untagged": False,
        "locked": False,
        "maximum_timeout": 28800,
        "access_level": "not_protected",
        "version": "15.4.0",
        "revision": "43b2dc3d",
        "platform": "linux",
        "architecture": "amd64",
        "contacted_at": "2022-10-13T01:49:58.577Z",
        "projects": [],
        "groups": []}

    with patch("server.gitlab.requests.get",
               return_value=MagicMock(status_code=200,
                                      json=MagicMock(return_value=details))) as get_mock:
        runner = runner_details(gitlab_url=url, private_token=private_token, runner_id=runner_id)
        assert runner == Runner(id=runner_id, description="keywords-gfx10-1",
                                active=True, paused=False, online=True,
                                tag_list=["TAG1", "TAG2"], maximum_timeout=28800)
        get_mock.assert_called_with(f"{url}/api/v4/runners/{runner_id}",
                                    headers={"PRIVATE-TOKEN": private_token})

    with patch("server.gitlab.requests.get", return_value=MagicMock(status_code=401)) as get_mock:
        assert runner_details(gitlab_url=url, private_token=private_token, runner_id=runner_id) is None


def test_generate_runner_config():
    template_data = "data"

    with patch("server.gitlab.Template") as template_mock:
        with patch("builtins.open", mock_open(read_data=template_data)) as mock_file:
            with patch("server.gitlab.psutil") as psutil_mock:
                psutil_mock.cpu_count.return_value = 4
                psutil_mock.virtual_memory.return_value.total = 1e9

                mars_db = MagicMock()

                generate_runner_config(mars_db)

                template_mock.assert_called_with(template_data)
                template_mock.return_value.render.assert_called_with(config=config, mars_db=mars_db,
                                                                     cpu_count=4, ram_total_MB=1000)

                mock_file.return_value.write.assert_called_with(template_mock.return_value.render.return_value)


def test_gitlab_runner_config_template():
    pdus = {
        "apc": {
            "driver": "apc_drv",
            "config": {}
        }
    }

    duts = {
        "de:ad:be:ef:ca:fe": {
            "base_name": "gfx9",
            "tags": ["tag1", "tag2"],
            "ip_address": "10.42.0.1",
            "local_tty_device": "ttyS0",
            "gitlab": {
                "freedesktop": {
                    "token": "mytoken",
                    "exposed": True
                }
            },
            "pdu": "apc",
            "pdu_port_id": 0,
            "pdu_off_delay": 30,
            "ready_for_service": True,
            "is_retired": False
        }
    }

    gitlab = {
        "freedesktop": {
            "url": "https://gitlab.freedesktop.org/",
            "registration_token": "<token>",
            "expose_runners": True,
            "maximum_timeout": 43200,
            "gateway_runner": {
                "token": "<token>",
                "exposed": True
            }
        }
    }

    mars_db = MarsDB(pdus=pdus, duts=duts, gitlab=gitlab)

    with open(config.GITLAB_CONF_TEMPLATE_FILE) as f:
        template = f.read()

    # Single board computers
    params = {
        "config": config,
        "mars_db": mars_db,
        "cpu_count": 4,
        "ram_total_MB": 4096
    }
    config_toml = str(Template(template).render(**params))
    assert 'limit = 1' in config_toml
    assert 'cpus = 4  # Gateway runner CPU allocation'
    assert 'memory = "2048 MB" # Gateway runner memory allocation' in config_toml
    assert 'memory_swap = "2457 MB" # Gateway runner memory hard limit' in config_toml
    assert 'memory_reservation = "1024 MB" # Gateway runner memory soft limit' in config_toml

    # Recycled x86 computer
    params = {
        "config": config,
        "mars_db": mars_db,
        "cpu_count": 8,
        "ram_total_MB": 16384
    }
    config_toml = str(Template(template).render(**params))
    assert 'limit = 1' in config_toml
    assert 'cpus = 8  # Gateway runner CPU allocation'
    assert 'memory = "12288 MB" # Gateway runner memory allocation' in config_toml
    assert 'memory_swap = "14745 MB" # Gateway runner memory hard limit' in config_toml
    assert 'memory_reservation = "6144 MB" # Gateway runner memory soft limit' in config_toml

    # Beefy server-class computer
    params = {
        "config": config,
        "mars_db": mars_db,
        "cpu_count": 32,
        "ram_total_MB": 131072
    }
    config_toml = str(Template(template).render(**params))
    assert 'limit = 4' in config_toml
    assert 'cpus = 32  # Gateway runner CPU allocation'
    assert 'memory = "126976 MB" # Gateway runner memory allocation' in config_toml
    assert 'memory_swap = "152371 MB" # Gateway runner memory hard limit' in config_toml
    assert 'memory_reservation = "63488 MB" # Gateway runner memory soft limit' in config_toml


@patch('platform.machine', return_value="x86_64")
@patch('psutil.cpu_count', return_value=4)
@patch('psutil.virtual_memory', return_value=MagicMock(total=3544967296))
def test_generate_gateway_runner_tags__low_power(vm_mock, cpu_count_mock, platform_mock):
    config.FARM_NAME = 'helloworld'
    assert generate_gateway_runner_tags() == {
        'helloworld-gateway', 'CI-gateway', 'cpu:arch:x86_64',
        'cpu:cores:4', 'mem:size:4GiB'
    }


@patch('platform.machine', return_value="arm64")
@patch('psutil.cpu_count', return_value=32)
@patch('psutil.virtual_memory', return_value=MagicMock(total=28359738368))
def test_generate_gateway_runner_tags__beast_machine(vm_mock, cpu_count_mock, platform_mock):
    config.FARM_NAME = 'helloworld'
    assert generate_gateway_runner_tags() == {
        'helloworld-gateway', 'CI-gateway', 'cpu:arch:arm64',
        'cpu:cores:32', 'cpu:cores:8+', 'cpu:cores:16+', 'cpu:cores:32+',
        'mem:size:32GiB', 'mem:size:8+GiB', 'mem:size:32+GiB'
    }



from copy import copy
from datetime import timedelta
from threading import Event
from unittest.mock import MagicMock, patch, PropertyMock, call

from server.pdu.daemon import AsyncPDU, AsyncPDUState, PDUDaemon


def init_mocked_pdu_ports(pdu_ports):
    pdu_ports.unlock_event = Event()

    def mockedPorts():
        pdu_ports.unlock_event.wait()
        return []

    pdu_ports.side_effect = mockedPorts


def test_AsyncPDU__eq():
    model_name = "dummy"
    pdu_name = "MyPDU"
    config = {"ports": ['P1', 'P2', 'P3']}
    reserved_port_ids = ["P1", "P2"]
    polling_interval = timedelta(seconds=1)

    pdu = AsyncPDU(model_name, pdu_name, config, reserved_port_ids, polling_interval)

    assert pdu == AsyncPDU(model_name, pdu_name, config, reserved_port_ids, polling_interval)
    assert pdu != AsyncPDU("vpdu", pdu_name, config, reserved_port_ids, polling_interval)
    assert pdu != AsyncPDU(model_name, "alternate name", config, reserved_port_ids, polling_interval)
    assert pdu != AsyncPDU(model_name, pdu_name, {"key": "value"}, reserved_port_ids, polling_interval)
    assert pdu != AsyncPDU(model_name, pdu_name, config, {'2', 'P2'}, polling_interval)
    assert pdu == AsyncPDU(model_name, pdu_name, config, reserved_port_ids, timedelta(seconds=2))


def test_AsyncPDU__get_port_by_id():
    pdu = AsyncPDU("dummy", "MyPDU", {"ports": ['P1', 'P2', 'P3']})

    pdu.started_event = MagicMock(is_set=MagicMock(return_value=False))
    pdu.ports = [MagicMock(port_id=0), MagicMock(port_id=1), MagicMock(port_id=2)]

    assert pdu.get_port_by_id(1) == pdu.ports[1]
    pdu.started_event.wait.assert_called_with(None)

    assert pdu.get_port_by_id(0, timeout=3) == pdu.ports[0]
    pdu.started_event.wait.assert_called_with(3)

    assert pdu.get_port_by_id(42) is None


@patch('server.pdu.drivers.dummy.DummyPDU.ports', new_callable=PropertyMock)
def test_AsyncPDU__run_keeps_going_when_failing_driver_setup(dummyPduPortsMock):
    dummyPduPortsMock.side_effect = ValueError("Custom error")

    pdu = AsyncPDU("dummy", "MyPDU", {"ports": ['P1', 'P2', 'P3']})

    pdu.stop_event = MagicMock(is_set=MagicMock(side_effect=[False, False, True]))

    pdu.start()
    pdu.join()

    assert not pdu.started_event.is_set()
    assert pdu.stop_event.wait.mock_calls == [call(15), call(15)]
    assert pdu.error == "Custom error"


@patch('server.pdu.drivers.dummy.DummyPDU.ports', new_callable=PropertyMock)
def test_AsyncPDU__run_recreate_driver_after_multiple_poll_fails(dummyPduPortsMock):
    pdu = AsyncPDU("dummy", "MyPDU", {"ports": ['P1', 'P2', 'P3']})

    expected_count = 10
    dummyPduPortsMock.side_effect = [[]] + expected_count * [ValueError("poll")] + [ValueError("final crash")]
    pdu.stop_event = MagicMock(is_set=MagicMock(side_effect=(1 + expected_count + 1) * [False] + [True]))

    pdu.start()
    pdu.join()

    assert not pdu.started_event.is_set()
    assert pdu.stop_event.is_set.call_count == 1 + expected_count + 1 + 1
    assert pdu.stop_event.wait.call_count == expected_count

    assert pdu.error == "final crash"


@patch('server.pdu.drivers.dummy.DummyPDU.ports', new_callable=PropertyMock)
def test_AsyncPDU__state(dummyPduPortsMock):
    # Set up the mock PDU to wait for a signal before returning the list of ports
    init_mocked_pdu_ports(dummyPduPortsMock)

    pdu = AsyncPDU("dummy", "MyPDU", {"ports": ['P1', 'P2', 'P3']})

    assert pdu.state == AsyncPDUState.CREATED

    # Test that we ignore any error set until the thread starts
    pdu.error = "Hello world"
    assert pdu.state == AsyncPDUState.CREATED
    pdu.error = None

    try:
        pdu.start()
        assert pdu.state == AsyncPDUState.INITIALIZING

        # Release the dummy PDUs ports signal to let the PDU finish initialization
        dummyPduPortsMock.unlock_event.set()
        pdu.started_event.wait()
        assert pdu.state == AsyncPDUState.OK

        pdu.error = "Hello world"
        assert pdu.state == AsyncPDUState.ERROR
        pdu.error = None

        pdu.stop()
        assert pdu.state == AsyncPDUState.STOPPED

        # Test that we ignore any error set after the thread has stopped
        pdu.error = "Hello world"
        assert pdu.state == AsyncPDUState.STOPPED
        pdu.error = None
    finally:
        pdu.stop()


def test_AsyncPDU__run_until_stop_event_set():
    pdu = AsyncPDU("dummy", "MyPDU", {"ports": ['P1', 'P2', 'P3']})
    pdu.stop_event = MagicMock(is_set=MagicMock(side_effect=[False, False, False, True, True]))

    pdu.run()

    assert pdu.stop_event.is_set.call_count == 5
    assert pdu.stop_event.wait.call_count == 2
    pdu.stop_event.wait.assert_called_with(1)


def test_AsyncPDU__stop_sets_stop_event_and_wait():
    pdu = AsyncPDU("dummy", "MyPDU", {"ports": ['P1', 'P2', 'P3']})

    pdu.stop_event = MagicMock()
    pdu.join = MagicMock()

    pdu.stop()

    pdu.stop_event.set.assert_called_once_with()
    pdu.join.assert_called_once_with()


@patch('server.pdu.drivers.dummy.DummyPDU')
def test_PDUDaemon__get_or_create(dummyPDUMock):
    daemon = PDUDaemon()

    try:
        # Make sure that by default we return a previously-created PDU, without updating *any* field
        ref = daemon.get_or_create("dummy", "MyPDU", {"ports": ['P1', 'P2', 'P3']})
        assert ref.polling_interval == timedelta(seconds=1)
        assert daemon.get_or_create("dummy", "MyPDU", {"ports": ['P2', 'P3', 'P4']},
                                    polling_interval=timedelta(seconds=5)) == ref
        assert ref.polling_interval == timedelta(seconds=1)

        # Ensure that the polling interval gets updated when the pdu would otherwise be the same
        assert ref.polling_interval == timedelta(seconds=1)
        assert daemon.get_or_create("dummy", "MyPDU", {"ports": ['P1', 'P2', 'P3']},
                                    update_if_existing=True, polling_interval=timedelta(seconds=5)) == ref
        assert ref.polling_interval == timedelta(seconds=5)

        # Make sure that updating the PDU's parameters lead to the creation of a new PDU and stops the old one
        assert ref.state == AsyncPDUState.OK
        assert daemon.get_or_create("dummy", "MyPDU", {"ports": ['P2', 'P3', 'P4']}, update_if_existing=True) != ref
        assert ref.state == AsyncPDUState.STOPPED
    finally:
        daemon.stop()


@patch('server.pdu.drivers.dummy.DummyPDU')
def test_PDUDaemon__unregister_pdu(dummyPDUMock):
    daemon = PDUDaemon()

    try:
        # Make sure that we can remove an existing PDU
        daemon.get_or_create("dummy", "MyPDU", {"ports": ['P1', 'P2', 'P3']})
        daemon.unregister_pdu("MyPDU")

        # Make sure that removing unknown PDUs doesn't raise
        daemon.unregister_pdu("MyPDU")
    finally:
        daemon.stop()


def test_PDUDaemon__stop():
    daemon = PDUDaemon()

    pdus = {"pdu1": MagicMock(), "pdu2": MagicMock()}
    daemon.pdus = copy(pdus)

    daemon.stop(wait=False)

    assert len(daemon.pdus) == 0
    pdus["pdu1"].stop.assert_called_once_with(wait=False)
    pdus["pdu2"].stop.assert_called_once_with(wait=False)



from unittest.mock import MagicMock, patch

from server.pdu import PDUState, PDUPort
from server.pdu.drivers.tasmota import TasmotaPDU


class TasmotaShellyDeviceMock:
    def get(url):
        url_prefix = "http://127.0.0.1"
        if url == f"{url_prefix}/cm?cmnd=Status":
            ret = {
                "Status": {
                    "Module": 0,
                    "DeviceName": "ShellyPlugSTasmota",
                    "FriendlyName": ["ShellyPlugSTasmota"],
                    "Topic": "tasmota_B4A75B",
                    "ButtonTopic": "0",
                    "Power": 0,
                    "PowerOnState": 3,
                    "LedState": 1,
                    "LedMask": "FFFF",
                    "SaveData": 1,
                    "SaveState": 1,
                    "SwitchTopic": "0",
                    "SwitchMode": [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                    "ButtonRetain": 0,
                    "SwitchRetain": 0,
                    "SensorRetain": 0,
                    "PowerRetain": 0,
                    "InfoRetain": 0,
                    "StateRetain": 0,
                    "StatusRetain": 0
                }
            }
        elif url == f"{url_prefix}/cm?cmnd=Power1":
            ret = {"POWER": "OFF"}
        elif url == f"{url_prefix}/cm?cmnd=Power1%20ON":
            ret = {"POWER": "ON"}
        else:  # pragma: nocover
            raise ValueError(f"Unexpected URL '{url}'")

        return MagicMock(json=MagicMock(return_value=ret))


@patch('server.pdu.drivers.tasmota.TasmotaPDU.requests_retry_session', TasmotaShellyDeviceMock)
def test_tasmota_shelly():
    pdu = TasmotaPDU('MyPDU', config={'hostname': '127.0.0.1'})

    assert pdu.num_ports == 1

    assert pdu.ports == [PDUPort(pdu=pdu, port_id="0", label=None)]
    assert pdu.get_port_state("0") == PDUState.OFF
    assert pdu.set_port_state("0", PDUState.ON)


class TasmotaGosundP1DeviceMock:
    def get(url):
        url_prefix = "http://127.0.0.1"
        if url == f"{url_prefix}/cm?cmnd=Status":
            ret = {
                "Status": {
                    "Module": 0,
                    "DeviceName": "Tasmota",
                    "FriendlyName": ["Tasmota", "", "", ""],
                    "Topic": "tasmota_E9D2B7",
                    "ButtonTopic": "0",
                    "Power": 0,
                    "PowerOnState": 3,
                    "LedState": 1,
                    "LedMask": "FFFF",
                    "SaveData": 1,
                    "SaveState": 1,
                    "SwitchTopic": "0",
                    "SwitchMode": [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                    "ButtonRetain": 0,
                    "SwitchRetain": 0,
                    "SensorRetain": 0,
                    "PowerRetain": 0,
                    "InfoRetain": 0,
                    "StateRetain": 0,
                    "StatusRetain": 0
                }
            }
        elif url == f"{url_prefix}/cm?cmnd=Power2":
            ret = {"POWER2": "OFF"}
        elif url == f"{url_prefix}/cm?cmnd=Power2%20ON":
            ret = {"POWER2": "ON"}
        else:  # pragma: nocover
            raise ValueError(f"Unexpected URL '{url}'")

        return MagicMock(json=MagicMock(return_value=ret))


@patch('server.pdu.drivers.tasmota.TasmotaPDU.requests_retry_session', TasmotaGosundP1DeviceMock)
def test_tasmota_gosund_p1():
    pdu = TasmotaPDU('MyPDU', config={'hostname': '127.0.0.1'})

    assert pdu.num_ports == 4
    assert pdu.get_port_state("1") == PDUState.OFF
    assert pdu.set_port_state("1", PDUState.ON)



from dataclasses import fields
from unittest.mock import MagicMock

from server.boots_db import USBMatcher, FastbootDeviceMatcher, BootsDB
from server import config

usb_device_fields = dict(idVendor=0x4242, idProduct=0x1234, iManufacturer="MuPuF TMI",
                         iProduct="CI-Tron", iSerialNumber="serialno")
usb_device = MagicMock(**usb_device_fields)


def test_USBMatcher():
    # Make sure the list of fields is up to date
    assert set(usb_device_fields.keys()) == {f.name for f in fields(USBMatcher)}

    # Empty matcher matches everything
    assert USBMatcher().matches(usb_device)

    # Single field match
    assert USBMatcher(iManufacturer=usb_device.iManufacturer).matches(usb_device)

    # Full match with direct values
    assert USBMatcher(idVendor=usb_device.idVendor, idProduct=usb_device.idProduct,
                      iManufacturer=usb_device.iManufacturer, iProduct=usb_device.iProduct,
                      iSerialNumber=usb_device.iSerialNumber).matches(usb_device)

    # Full match with lists of acceptable values
    assert USBMatcher(idVendor=[usb_device.idVendor, 0x4343],
                      idProduct=[0x4321, usb_device.idProduct],
                      iManufacturer=["Someone else", usb_device.iManufacturer],
                      iProduct=["OtherProduct", usb_device.iProduct],
                      iSerialNumber=[usb_device.iSerialNumber, "OtherSerial"]).matches(usb_device)

    # Make sure that every field is important for the match
    for key, value in usb_device_fields.items():
        new_fields = dict(usb_device_fields)

        if isinstance(value, int):
            new_fields[key] += 1
        elif isinstance(value, str):
            new_fields[key] += "-"

        assert not USBMatcher(**new_fields).matches(usb_device)


def test_FastbootDeviceMatcher():
    usb_matcher = USBMatcher(**usb_device_fields)
    full_variables = {"product": "myproduct", "manufacturer": "MuPuF TMI", "serialno": "123456"}

    # Empty means matches everything!
    assert FastbootDeviceMatcher().matches(MagicMock())

    # USB match only
    assert FastbootDeviceMatcher(usb=usb_matcher).matches(MagicMock(device=usb_device))
    assert not FastbootDeviceMatcher(usb=USBMatcher(idVendor=0x1234)).matches(MagicMock(device=usb_device))

    # Make sure that no variables set means a match!
    assert FastbootDeviceMatcher(variables={}).matches(MagicMock(variables=full_variables))

    # Try adding variables and make sure it still matches
    match_variables = {}
    for var, value in full_variables.items():
        match_variables[var] = value
        assert FastbootDeviceMatcher(variables=match_variables).matches(MagicMock(variables=full_variables))

    # Ensure that missing variables do not count as a match
    assert not FastbootDeviceMatcher(variables={"missing": "value"}).matches(MagicMock(variables={}))

    # Ensure that variables with different values do not count as a match
    assert not FastbootDeviceMatcher(variables={"key": "value"}).matches(MagicMock(variables={"key": "othervalue"}))


def test_BootsDB_render():
    # Make sure we can parse the default database, without any parameter
    BootsDB.from_path(config.BOOTS_DB_FILE)



import socket
import os

from collections import defaultdict


def named_fds():
    # Ignore any FDs that were not meant for us
    if os.environ.get('LISTEN_PID', None) != str(os.getpid()):
        return {}

    try:
        listen_fds_nr = int(os.getenv("LISTEN_FDS"))
    except Exception:
        return {}

    FIRST_SOCKET_FD = 3
    fds = [FIRST_SOCKET_FD + i for i in range(listen_fds_nr)]
    names = os.getenv("LISTEN_FDNAMES", "").split(":")

    named_fds = defaultdict(list)
    for i, fd in enumerate(fds):
        name = names[i] if i < len(names) and names[i] else None
        named_fds[name].append(fd)

    return named_fds


def get_sockets_by_name(name=None, family=socket.AF_INET, type=socket.SOCK_STREAM, proto=0):
    fds = named_fds().get(name, [])
    return [socket.socket(family, type, proto, fd) for fd in fds]



from easysnmp import Session
import easysnmp.exceptions
from .. import logger, PDU, PDUPort, PDUState
from functools import cached_property
from typing import Dict

import random
import time


def _is_int(s):
    try:
        s = int(s)
        return True
    except ValueError:
        return False


def retry_on_known_errors(func):
    def retry(*args, **kwargs):
        retries = 3

        for i in range(retries):
            try:
                return func(*args, **kwargs)
            except (SystemError, easysnmp.exceptions.EasySNMPError) as e:
                logger.warning(f"Caught the error '{str(e)}', retrying ({i+1}/{retries})")
                # Wait 1 second, plus a random [0,1] second delay to reduce the chances of concurrent requests
                time.sleep(1 + random.random())
                continue

        raise ValueError(f"The function {func} failed {retries} times in a row")

    return retry


class SnmpPDU(PDU):
    oid_enterprise = '1.3.6.1.4.1'

    @property
    def __port_labels(self):
        try:
            return [x.value for x in self.session.walk(self.outlet_labels_oid)]
        except SystemError as e:
            raise ValueError(f"The snmp_walk() call failed with the following error: {e}")

    def __init__(self, name, config, reserved_port_ids=[]):
        super().__init__(name, config, reserved_port_ids)

        assert self.system_id
        assert self.outlet_labels
        assert self.outlet_status

        if not hasattr(self, 'outlet_ctrl'):
            # Some PDUs offer a RW status tree, others require a separate
            # tree for writes. Default to the seemingly more common case
            # of a RW tree.
            self.outlet_ctrl = self.outlet_status

        # FIXME: The UNKNOWN status is a bit of an odd one, not all PDUs expose such a concept.
        assert self.state_mapping.keys() == set([PDUState.ON, PDUState.OFF, PDUState.REBOOT])
        if not hasattr(self, 'inverse_state_mapping'):
            self.inverse_state_mapping: Dict[int, PDUState] = \
                dict([(value, key) for key, value in self.state_mapping.items()])
        else:
            assert self.inverse_state_mapping.keys() == set([PDUState.ON, PDUState.OFF, PDUState.REBOOT])

        # Validate the configuration by getting the list of ports
        self._ports = []
        self.ports

    @cached_property
    def session(self):
        config = self.config

        if 'hostname' not in config:
            raise ValueError('SnmpPDU requires a "hostname" configuration key')

        version = config.get('version', 1)
        if version in [1, 2]:
            session = Session(hostname=config['hostname'], community=config.get('community', 'private'),
                              version=version)
        elif version == 3:
            # Only keep the keys
            supported_keys = {'hostname', 'security_username', 'privacy_protocol', 'privacy_password',
                              'auth_protocol', 'auth_password', 'context_engine_id', 'security_engine_id', 'version'}
            session_cfg = {k: v for k, v in config.items() if k in supported_keys}

            auth_protocol = session_cfg.get('auth_protocol')
            privacy_protocol = session_cfg.get('privacy_protocol')
            if auth_protocol is not None and privacy_protocol is not None:
                session_cfg['security_level'] = 'auth_with_privacy'
            elif auth_protocol is not None and privacy_protocol is None:
                session_cfg['security_level'] = 'auth_without_privacy'
            elif auth_protocol is None and privacy_protocol is None:
                session_cfg['security_level'] = 'no_auth_or_privacy'
            else:
                raise ValueError("Unsupported security level: Can't have a privacy protocol with no auth protocol")

            session = Session(**session_cfg)
        else:
            raise ValueError(f"SNMP version {version} is unsupported")

        # Validate the configuration by reading the state of the first port
        session.get(self.outlet_status_oid(1))

        return session

    @property
    def outlet_system_id(self):
        return f'{self.oid_enterprise}.{self.system_id}'

    @property
    def outlet_labels_oid(self):
        return f'{self.outlet_system_id}.{self.outlet_labels}'

    def outlet_status_oid(self, port_id):
        assert isinstance(port_id, int)
        return f'{self.outlet_system_id}.{self.outlet_status}.{port_id}'

    def outlet_ctrl_oid(self, port_id: int):
        assert isinstance(port_id, int)
        return f'{self.outlet_system_id}.{self.outlet_ctrl}.{port_id}'

    @property
    def ports(self):
        labels = self.__port_labels
        for i, label in enumerate(labels):
            if len(self._ports) <= i:
                port = PDUPort(pdu=self, port_id=i+1)
                self._ports.append(port)
            else:
                port = self._ports[i]

            # Update the label
            port.label = labels[i]

        # Truncate the list of ports if it shrank
        self._ports = self._ports[0:len(labels)]

        return self._ports

    def _port_spec_to_int(self, port_spec):
        if _is_int(port_spec):
            return port_spec
        else:
            for port in self.ports:
                if port.label == port_spec:
                    return port.port_id
            raise ValueError(
                f"{port_spec} can not be interpreted as a valid port")

    @retry_on_known_errors
    def set_port_state(self, port_spec, state):
        SNMP_INTEGER_TYPE = 'i'

        port_id = self._port_spec_to_int(port_spec)
        logger.debug('setting OID %s to state %s with value %d',
                     self.outlet_ctrl_oid(port_id),
                     state,
                     self.state_mapping[state])
        ret = self.session.set(self.outlet_ctrl_oid(port_id),
                               self.state_mapping[state],
                               SNMP_INTEGER_TYPE)

        if self.state_transition_delay_seconds is not None:
            logger.debug("Enforcing %s seconds of delay for state change", self.state_transition_delay_seconds)
            # TODO: keep track of state changes to avoid a forced sleep.
            # TODO: Polling for the state change would be better in general.
            # The root cause of this is because PDUs maintain their
            # own configurables how long to delay between
            # transitions, we should probably control that via SNMP,
            # as well.
            time.sleep(self.state_transition_delay_seconds)

        return ret

    @retry_on_known_errors
    def get_port_state(self, port_spec):
        port_id = self._port_spec_to_int(port_spec)
        vs = self.session.get(self.outlet_status_oid(port_id))
        value = int(vs.value)
        logger.debug('retrieved OID %s with value %d, maps to state %s',
                     self.outlet_status_oid(port_id),
                     value,
                     self.inverse_state_mapping[value])
        return self.inverse_state_mapping[value]

    def __eq__(self, other):
        return not any([
            getattr(self, attr, None) != getattr(other, attr, None)
            for attr in ["name",
                         "config",
                         "system_id",
                         "outlet_labels",
                         "outlet_status",
                         "outlet_ctrl",
                         "state_mapping",
                         "inverse_state_mapping"]])


class ManualSnmpPDU(SnmpPDU):
    def __init__(self, name, config, reserved_port_ids=[]):
        if inverse_state_mapping := config.get('inverse_state_mapping'):
            self.inverse_state_mapping = self.__generate_state_mapping(inverse_state_mapping)

        super().__init__(name, config, reserved_port_ids)

    @property
    def system_id(self):
        return self.config['system_id']

    @property
    def outlet_labels(self):
        return self.config['outlet_labels']

    @property
    def outlet_status(self):
        return self.config['outlet_status']

    # Some PDUs offer a RW status tree, others require a separate
    # tree for writes. Default to the seemingly more common case
    # of a RW tree.
    @property
    def outlet_ctrl(self):
        return self.config.get('outlet_ctrl', self.outlet_status)

    def __generate_state_mapping(self, d):
        state_mapping = dict()

        for state, internal_value in d.items():
            v = int(internal_value)
            if state.lower() == "on":
                state_mapping[PDUState.ON] = v
            elif state.lower() == "off":
                state_mapping[PDUState.OFF] = v
            elif state.lower() == "reboot":
                state_mapping[PDUState.REBOOT] = v
                # Unknown deliberately excluded.

        return state_mapping

    @property
    def state_mapping(self):
        return self.__generate_state_mapping(self.config.get('state_mapping', {}))



from unittest.mock import MagicMock, PropertyMock, patch, call
import io

from server.android.fastbootd import FastbootDevice, Fastbootd

import pytest
import usb.util


@patch("server.android.fastbootd.usb.util.find_descriptor")
@patch("server.android.fastbootd.usb.util.endpoint_type")
@patch("server.android.fastbootd.usb.util.endpoint_direction")
def FastbootDeviceMock(endpoint_direction_mock, endpoint_type_mock, find_descriptor_mock, *args, **kwargs):
    device = MagicMock()
    cfg = MagicMock()
    intf = MagicMock()
    ep_write = MagicMock()
    ep_read = MagicMock()

    device.__iter__.return_value = [cfg]
    find_descriptor_mock.return_value = intf
    intf.__iter__.return_value = [ep_read, ep_write]
    endpoint_type_mock.return_value = usb.util.ENDPOINT_TYPE_BULK
    endpoint_direction_mock.side_effect = [usb.util.ENDPOINT_IN, usb.util.ENDPOINT_OUT]

    fb_dev = FastbootDevice(device, *args, **kwargs)

    find_descriptor_mock.assert_called_once_with(cfg,
                                                 bInterfaceClass=0xff,
                                                 bInterfaceSubClass=0x42,
                                                 bInterfaceProtocol=0x03)
    endpoint_type_mock.assert_any_call(ep_write.bmAttributes)
    endpoint_type_mock.assert_any_call(ep_read.bmAttributes)
    endpoint_direction_mock.assert_any_call(ep_write.bEndpointAddress)
    endpoint_direction_mock.assert_any_call(ep_read.bEndpointAddress)

    return fb_dev


@patch("server.android.fastbootd.usb.core.find")
def test_FastbootDevice__from_serial(find_mock):
    dev0 = MagicMock()
    dev1 = MagicMock(serial_number="0123456789")
    dev2 = MagicMock(serial_number="9876543210")

    # Make the first device always raise an exception when accessing its serial number
    type(dev0).serial_number = PropertyMock(side_effect=ValueError)

    find_mock.return_value = [dev0, dev1, dev2]
    with patch("server.android.fastbootd.FastbootDevice.__init__", return_value=None) as init_mock:
        FastbootDevice.from_serial("9876543210", 42)
        find_mock.assert_called_once_with(find_all=True)
        init_mock.assert_called_once_with(dev2, write_chunk=42)

    with pytest.raises(ValueError):
        FastbootDevice.from_serial("serial", 42)


def test_FastbootDevice__basic_io__read():
    dev = FastbootDeviceMock()
    assert dev.read(1024, timeout=42) == dev.ep_read.read.return_value
    dev.ep_read.read.assert_called_once_with(1024, timeout=42)


def test_FastbootDevice__release():
    dev = FastbootDeviceMock()

    dev.device._ctx.release_all_interfaces.assert_not_called()
    dev.release()
    dev.device._ctx.release_all_interfaces.assert_called_once_with(dev.device)


def test_FastbootDevice__basic_io__write():
    dev = FastbootDeviceMock(write_chunk=2)

    dev.write(b"abcde", timeout=42) == dev.ep_read.read.return_value

    assert dev.ep_write.write.call_count == 3
    dev.ep_write.write.assert_has_calls([call(b"ab", timeout=42), call(b"cd", timeout=42), call(b"e", timeout=42)])


def test_FastbootDevice__runcmd__basic():
    dev = FastbootDeviceMock()
    dev.write = MagicMock()
    dev.read = MagicMock(side_effect=[b"INFOline1", b"INFOline2", b"INFOline3", b"OKAY"])

    dev.run_cmd("cmd", "arg", timeout=1)

    dev.write.assert_called_once_with(b"cmd:arg")
    assert dev.read.call_count == 4
    dev.read.assert_has_calls([call(64, timeout=1), call(64, timeout=1), call(64, timeout=1), call(64, timeout=1)])


def test_FastbootDevice__runcmd__fail():
    dev = FastbootDeviceMock()
    dev.read = MagicMock(return_value=b"FAILcmd failed!")

    with pytest.raises(ValueError):
        dev.run_cmd("cmd")


def test_FastbootDevice__runcmd__unwanted_status():
    dev = FastbootDeviceMock()
    dev.read = MagicMock(return_value=b"DATAcmd failed!")

    with pytest.raises(ValueError):
        dev.run_cmd("cmd", wanted_status="OKAY")


def test_FastbootDevice__getvar():
    dev = FastbootDeviceMock()

    # Missing variable
    dev.run_cmd = MagicMock(return_value=[(None, None)])
    assert dev.getvar("KEY") is None

    # Single variable
    dev.run_cmd = MagicMock(return_value=[(None, b"KEY:VAL"), (None, None)])
    assert dev.getvar("KEY") == "VAL"

    # Multiple variables
    dev.run_cmd = MagicMock(return_value=[(None, b"KEY:VAL"), (None, b"KEY2:VAL2"), (None, None)])
    assert dev.getvar("all") == {"KEY": "VAL", "KEY2": "VAL2"}


def test_FastbootDevice__variables():
    dev = FastbootDeviceMock()

    dev.getvar = MagicMock()
    assert dev.variables == dev.getvar.return_value
    dev.getvar.assert_called_once_with(b"all")


def test_FastbootDevice__upload():
    data = io.BytesIO(b"My data")

    dev = FastbootDeviceMock(write_chunk=2)
    dev.read = MagicMock(side_effect=[b"DATA00000007", b"OKAY"])

    dev.upload(data)

    dev.ep_write.write.assert_has_calls([call(b"My"), call(b" d"), call(b"at"), call(b"a")])


@patch("server.android.fastbootd.usb.core.find")
def test_FastbootDevice__boot(find_mock):
    dev = FastbootDeviceMock()

    dev.run_cmd = MagicMock()
    dev.boot()
    dev.run_cmd.assert_called_once_with(b"boot", timeout=15000)


def FastbootDeviceNewMock(cls, device, **kwargs):
    if device.raises:
        raise ValueError("Mock Error!")

    return MagicMock(device=device, **kwargs)


@patch("server.android.fastbootd.FastbootDevice.__new__", side_effect=FastbootDeviceNewMock)
def test_Fastbootd(fbdev_mock):
    def new_loop(found={}, added=set(), rmed=set()):
        # Mock the callbacks and make them assert to ensure the daemon keeps going no matter what
        daemon.new_device_found = MagicMock(side_effect=AssertionError)
        daemon.device_removed = MagicMock(side_effect=ValueError)

        with patch("server.android.fastbootd.usb.core.find", return_value=found) as find_mock:
            daemon.update_device_list()

            find_mock.assert_called_once_with(find_all=True)
            assert {call.args[0].device for call in daemon.new_device_found.call_args_list} == added
            assert {call.args[0].device for call in daemon.device_removed.call_args_list} == rmed

    daemon = Fastbootd()

    devs = [MagicMock(raises=False) for i in range(3)]
    non_fastbootd_dev = MagicMock(raises=True)

    # Make sure any assert from core find does not get propagated
    with patch("server.android.fastbootd.usb.core.find", side_effect=Exception):
        daemon.update_device_list()

    # First discovery
    new_loop(found=[devs[0], devs[1]], added={devs[0], devs[1]}, rmed=set())

    # No changes but order different
    new_loop(found=[devs[1], devs[0]], added=set(), rmed=set())

    # Remove one, add a new one
    new_loop(found=[devs[2], devs[0]], added={devs[2]}, rmed={devs[1]})

    # Add a non-fastboot device
    new_loop(found=[devs[2], non_fastbootd_dev, devs[0]], added=set(), rmed=set())



from unittest.mock import patch
import pytest

from server.pdu import PDUState
from server.pdu.drivers.devantech import DevantechPDU


class DevantechEth0020SocketNoPasswordMock:
    def __init__(self, *args, **kwargs):
        self._data_to_recv = None

    def setsockopt(self, level, optname, value):
        pass

    def connect(self, address):
        pass  # Simulate a successful connection

    def recv(self, bufsize, flags=0):
        assert (len(self._data_to_recv) == bufsize)
        return self._data_to_recv

    def sendall(self, data):
        if data == b'\x10':
            self._data_to_recv = b'\x15\x00\x00'
        elif data == b'\x24':
            self._data_to_recv = b'\x60\x80\x0c'
        elif data == b'\x7a':
            self._data_to_recv = b'\xff'
        pass

    def close(self):
        pass

    # Context manager protocol methods
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


@patch('socket.socket', new=DevantechEth0020SocketNoPasswordMock)
def test_devantech_eth0020_no_password_successful_connection():
    pdu = DevantechPDU("TestPDU", {})

    assert len(pdu.ports) == 20
    assert pdu.get_port_state(0) == PDUState.OFF
    assert pdu.get_port_state(1) == PDUState.OFF
    assert pdu.get_port_state(2) == PDUState.OFF
    assert pdu.get_port_state(3) == PDUState.OFF
    assert pdu.get_port_state(4) == PDUState.OFF
    assert pdu.get_port_state(5) == PDUState.ON
    assert pdu.get_port_state(6) == PDUState.ON
    assert pdu.get_port_state(7) == PDUState.OFF
    assert pdu.get_port_state(8) == PDUState.OFF
    assert pdu.get_port_state(9) == PDUState.OFF
    assert pdu.get_port_state(10) == PDUState.OFF
    assert pdu.get_port_state(11) == PDUState.OFF
    assert pdu.get_port_state(12) == PDUState.OFF
    assert pdu.get_port_state(13) == PDUState.OFF
    assert pdu.get_port_state(14) == PDUState.OFF
    assert pdu.get_port_state(15) == PDUState.ON
    assert pdu.get_port_state(16) == PDUState.OFF
    assert pdu.get_port_state(17) == PDUState.OFF
    assert pdu.get_port_state(18) == PDUState.ON
    assert pdu.get_port_state(19) == PDUState.ON


class DevantechEth008SocketPasswordMock:
    def __init__(self, *args, **kwargs):
        self._data_to_recv = None

    def setsockopt(self, level, optname, value):
        pass

    def connect(self, address):
        pass  # Simulate a successful connection

    def recv(self, bufsize, flags=0):
        assert (len(self._data_to_recv) == bufsize)
        return self._data_to_recv

    def sendall(self, data):
        if data == b'\x10':
            self._data_to_recv = b'\x13\x00\x00'
        elif data == b'\x24':
            self._data_to_recv = b'\x98'
        elif data == b'\x7a':
            self._data_to_recv = b'\x00'
        elif data.startswith(b'\x79'):
            # Verify password
            if data[1:] == b'top-secret':
                self._data_to_recv = b'\x01'
            else:
                self._data_to_recv = b'\x02'
        pass

    def close(self):
        pass

    # Context manager protocol methods
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


@patch('socket.socket', new=DevantechEth008SocketPasswordMock)
def test_devantech_eth008_password_successful_connection():
    pdu = DevantechPDU("TestPDU", {"password": "top-secret"})

    assert len(pdu.ports) == 8
    assert pdu.get_port_state(0) == PDUState.OFF
    assert pdu.get_port_state(1) == PDUState.OFF
    assert pdu.get_port_state(2) == PDUState.OFF
    assert pdu.get_port_state(3) == PDUState.ON
    assert pdu.get_port_state(4) == PDUState.ON
    assert pdu.get_port_state(5) == PDUState.OFF
    assert pdu.get_port_state(6) == PDUState.OFF
    assert pdu.get_port_state(7) == PDUState.ON


@patch('socket.socket', new=DevantechEth008SocketPasswordMock)
def test_devantech_eth008_wrong_password():
    with pytest.raises(ValueError) as excinfo:
        DevantechPDU("TestPDU", {"password": "not-top-secret"})

    assert str(excinfo.value) == "wrong password"


class DevantechUnknownIdSocketMock:
    def __init__(self, *args, **kwargs):
        self._data_to_recv = None

    def setsockopt(self, level, optname, value):
        pass

    def connect(self, address):
        pass  # Simulate a successful connection

    def recv(self, bufsize, flags=0):
        assert (len(self._data_to_recv) == bufsize)
        return self._data_to_recv

    def sendall(self, data):
        if data == b'\x10':
            self._data_to_recv = b'\x42\x00\x00'
        elif data == b'\x7a':
            self._data_to_recv = b'\xff'
        pass

    def close(self):
        pass

    # Context manager protocol methods
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


@patch('socket.socket', new=DevantechUnknownIdSocketMock)
def test_devantech_unknown_id():
    with pytest.raises(ValueError) as excinfo:
        DevantechPDU("TestPDU", {})

    assert str(excinfo.value) == "not supported module id found"


class DevantechSetPortStateFailSocketMock:
    def __init__(self, *args, **kwargs):
        self._data_to_recv = None

    def setsockopt(self, level, optname, value):
        pass

    def connect(self, address):
        pass  # Simulate a successful connection

    def recv(self, bufsize, flags=0):
        assert (len(self._data_to_recv) == bufsize)
        return self._data_to_recv

    def sendall(self, data):
        if data == b'\x10':
            self._data_to_recv = b'\x13\x00\x00'
        elif data.startswith(b'\x20'):
            self._data_to_recv = b'\x01'  # failed to set port state
        elif data == b'\x7a':
            self._data_to_recv = b'\xff'
        pass

    def close(self):
        pass

    # Context manager protocol methods
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


@patch('socket.socket', new=DevantechSetPortStateFailSocketMock)
def test_devantech_set_port_state_fail():
    pdu = DevantechPDU("TestPDU", {})

    with pytest.raises(ValueError) as excinfo:
        pdu.set_port_state(0, PDUState.ON)

    assert str(excinfo.value) == "failed to set port state"


class DevantechGetPortStateFailSocketMock:
    def __init__(self, *args, **kwargs):
        self._data_to_recv = None

    def setsockopt(self, level, optname, value):
        pass

    def connect(self, address):
        pass  # Simulate a successful connection

    def recv(self, bufsize, flags=0):
        return self._data_to_recv

    def sendall(self, data):
        if data == b'\x10':
            self._data_to_recv = b'\x13\x00\x00'
        elif data.startswith(b'\x24'):
            self._data_to_recv = b''  # failed to get port state
        elif data == b'\x7a':
            self._data_to_recv = b'\xff'
        pass

    def close(self):
        pass

    # Context manager protocol methods
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


@patch('socket.socket', new=DevantechGetPortStateFailSocketMock)
def test_devantech_get_port_state_fail():
    pdu = DevantechPDU("TestPDU", {})

    with pytest.raises(ValueError) as excinfo:
        pdu.get_port_state(0)

    assert str(excinfo.value) == "we never received the expected amount of bytes"
